"""Unit test for saltcheck execution module"""

import os.path

import pytest

import salt.config
import salt.modules.saltcheck as saltcheck
import salt.syspaths as syspaths
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class SaltcheckTestCase(TestCase, LoaderModuleMockMixin):
    """
    TestCase for salt.modules.saltcheck module
    """

    def setup_loader_modules(self):
        # Setting the environment to be local
        local_opts = salt.config.minion_config(
            os.path.join(syspaths.CONFIG_DIR, "minion")
        )
        local_opts["file_client"] = "local"
        local_opts["conf_file"] = "/etc/salt/minion"
        patcher = patch("salt.config.minion_config", MagicMock(return_value=local_opts))
        patcher.start()
        self.addCleanup(patcher.stop)
        return {saltcheck: {"__opts__": local_opts}}

    @pytest.mark.slow_test
    def test_call_salt_command(self):
        """test simple test.echo module"""
        with patch.dict(
            saltcheck.__salt__,
            {
                "config.get": MagicMock(return_value=True),
                "sys.list_modules": MagicMock(return_value=["module1"]),
                "cp.cache_master": MagicMock(return_value=[True]),
            },
        ):
            sc_instance = saltcheck.SaltCheck()
            returned = sc_instance._call_salt_command(
                fun="test.echo", args=["hello"], kwargs=None
            )
            self.assertEqual(returned, "hello")

    @pytest.mark.slow_test
    def test_call_salt_command2(self):
        """test simple test.echo module again"""
        with patch.dict(
            saltcheck.__salt__,
            {
                "config.get": MagicMock(return_value=True),
                "sys.list_modules": MagicMock(return_value=["module1"]),
                "cp.cache_master": MagicMock(return_value=[True]),
            },
        ):
            sc_instance = saltcheck.SaltCheck()
            returned = sc_instance._call_salt_command(
                fun="test.echo", args=["hello"], kwargs=None
            )
            self.assertNotEqual(returned, "not-hello")

    def test_call_saltcheck_with_proxy(self):
        """test fail to load saltcheck module if proxy"""
        with patch.dict(
            saltcheck.__salt__,
            {
                "config.get": MagicMock(return_value=True),
                "sys.list_modules": MagicMock(return_value=["module1"]),
                "cp.cache_master": MagicMock(return_value=[True]),
            },
        ), patch("salt.utils.platform.is_proxy", MagicMock(return_value=[True])):
            ret, message = saltcheck.__virtual__()
            self.assertFalse(ret)
            self.assertEqual(
                message,
                "The saltcheck execution module failed to load: only available on"
                " minions.",
            )

    def test__assert_equal1(self):
        """test"""
        with patch.dict(
            saltcheck.__salt__,
            {
                "config.get": MagicMock(return_value=True),
                "cp.cache_master": MagicMock(return_value=[True]),
            },
        ):
            sc_instance = saltcheck.SaltCheck()
            aaa = {"a": 1, "b": 2}
            bbb = {"a": 1, "b": 2}
            mybool = sc_instance._SaltCheck__assert_equal(aaa, bbb)
            self.assertTrue(mybool)

    def test__assert_equal2(self):
        """test"""
        with patch.dict(
            saltcheck.__salt__,
            {
                "config.get": MagicMock(return_value=True),
                "cp.cache_master": MagicMock(return_value=[True]),
            },
        ):
            sc_instance = saltcheck.SaltCheck()
            mybool = sc_instance._SaltCheck__assert_equal(False, True)
            self.assertNotEqual(mybool, True)

    def test__assert_not_equal1(self):
        """test"""
        with patch.dict(
            saltcheck.__salt__,
            {
                "config.get": MagicMock(return_value=True),
                "cp.cache_master": MagicMock(return_value=[True]),
            },
        ):
            sc_instance = saltcheck.SaltCheck()
            aaa = {"a": 1, "b": 2}
            bbb = {"a": 1, "b": 2, "c": 3}
            mybool = sc_instance._SaltCheck__assert_not_equal(aaa, bbb)
            self.assertTrue(mybool)

    def test__assert_not_equal2(self):
        """test"""
        with patch.dict(
            saltcheck.__salt__,
            {
                "config.get": MagicMock(return_value=True),
                "cp.cache_master": MagicMock(return_value=[True]),
            },
        ):
            sc_instance = saltcheck.SaltCheck()
            aaa = {"a": 1, "b": 2}
            bbb = {"a": 1, "b": 2}
            mybool = sc_instance._SaltCheck__assert_not_equal(aaa, bbb)
            self.assertNotEqual(mybool, True)

    def test__assert_true1(self):
        """test"""
        with patch.dict(
            saltcheck.__salt__,
            {
                "config.get": MagicMock(return_value=True),
                "cp.cache_master": MagicMock(return_value=[True]),
            },
        ):
            sc_instance = saltcheck.SaltCheck()
            mybool = sc_instance._SaltCheck__assert_equal(True, True)
            self.assertTrue(mybool)

    def test__assert_true2(self):
        """test"""
        with patch.dict(
            saltcheck.__salt__,
            {
                "config.get": MagicMock(return_value=True),
                "cp.cache_master": MagicMock(return_value=[True]),
            },
        ):
            sc_instance = saltcheck.SaltCheck()
            mybool = sc_instance._SaltCheck__assert_equal(False, True)
            self.assertNotEqual(mybool, True)

    def test__assert_false1(self):
        """test"""
        with patch.dict(
            saltcheck.__salt__,
            {
                "config.get": MagicMock(return_value=True),
                "cp.cache_master": MagicMock(return_value=[True]),
            },
        ):
            sc_instance = saltcheck.SaltCheck()
            mybool = sc_instance._SaltCheck__assert_false(False)
            self.assertTrue(mybool)

    def test__assert_false2(self):
        """test"""
        with patch.dict(
            saltcheck.__salt__,
            {
                "config.get": MagicMock(return_value=True),
                "cp.cache_master": MagicMock(return_value=[True]),
            },
        ):
            sc_instance = saltcheck.SaltCheck()
            mybool = sc_instance._SaltCheck__assert_false(True)
            self.assertNotEqual(mybool, True)

    def test__assert_in1(self):
        """test"""
        with patch.dict(
            saltcheck.__salt__,
            {
                "config.get": MagicMock(return_value=True),
                "cp.cache_master": MagicMock(return_value=[True]),
            },
        ):
            sc_instance = saltcheck.SaltCheck()
            aaa = "bob"
            mylist = ["alice", "bob", "charles", "dana"]
            mybool = sc_instance._SaltCheck__assert_in(aaa, mylist)
            self.assertTrue(mybool, True)

    def test__assert_in2(self):
        """test"""
        with patch.dict(
            saltcheck.__salt__,
            {
                "config.get": MagicMock(return_value=True),
                "cp.cache_master": MagicMock(return_value=[True]),
            },
        ):
            sc_instance = saltcheck.SaltCheck()
            aaa = "elaine"
            mylist = ["alice", "bob", "charles", "dana"]
            mybool = sc_instance._SaltCheck__assert_in(aaa, mylist)
            self.assertNotEqual(mybool, True)

    def test__assert_not_in1(self):
        """test"""
        with patch.dict(
            saltcheck.__salt__,
            {
                "config.get": MagicMock(return_value=True),
                "cp.cache_master": MagicMock(return_value=[True]),
            },
        ):
            sc_instance = saltcheck.SaltCheck()
            aaa = "elaine"
            mylist = ["alice", "bob", "charles", "dana"]
            mybool = sc_instance._SaltCheck__assert_not_in(aaa, mylist)
            self.assertTrue(mybool, True)

    def test__assert_not_in2(self):
        """test"""
        with patch.dict(
            saltcheck.__salt__,
            {
                "config.get": MagicMock(return_value=True),
                "cp.cache_master": MagicMock(return_value=[True]),
            },
        ):
            sc_instance = saltcheck.SaltCheck()
            aaa = "bob"
            mylist = ["alice", "bob", "charles", "dana"]
            mybool = sc_instance._SaltCheck__assert_not_in(aaa, mylist)
            self.assertNotEqual(mybool, True)

    def test__assert_greater1(self):
        """test"""
        with patch.dict(
            saltcheck.__salt__,
            {
                "config.get": MagicMock(return_value=True),
                "cp.cache_master": MagicMock(return_value=[True]),
            },
        ):
            sc_instance = saltcheck.SaltCheck()
            aaa = 110
            bbb = 100
            mybool = sc_instance._SaltCheck__assert_greater(aaa, bbb)
            self.assertTrue(mybool, True)

    def test__assert_greater2(self):
        """test"""
        with patch.dict(
            saltcheck.__salt__,
            {
                "config.get": MagicMock(return_value=True),
                "cp.cache_master": MagicMock(return_value=[True]),
            },
        ):
            sc_instance = saltcheck.SaltCheck()
            aaa = 100
            bbb = 110
            mybool = sc_instance._SaltCheck__assert_greater(aaa, bbb)
            self.assertNotEqual(mybool, True)

    def test__assert_greater3(self):
        """test"""
        with patch.dict(
            saltcheck.__salt__,
            {
                "config.get": MagicMock(return_value=True),
                "cp.cache_master": MagicMock(return_value=[True]),
            },
        ):
            sc_instance = saltcheck.SaltCheck()
            aaa = 100
            bbb = 100
            mybool = sc_instance._SaltCheck__assert_greater(aaa, bbb)
            self.assertNotEqual(mybool, True)

    def test__assert_greater_equal1(self):
        """test"""
        with patch.dict(
            saltcheck.__salt__,
            {
                "config.get": MagicMock(return_value=True),
                "cp.cache_master": MagicMock(return_value=[True]),
            },
        ):
            sc_instance = saltcheck.SaltCheck()
            aaa = 110
            bbb = 100
            mybool = sc_instance._SaltCheck__assert_greater_equal(aaa, bbb)
            self.assertTrue(mybool, True)

    def test__assert_greater_equal2(self):
        """test"""
        with patch.dict(
            saltcheck.__salt__,
            {
                "config.get": MagicMock(return_value=True),
                "cp.cache_master": MagicMock(return_value=[True]),
            },
        ):
            sc_instance = saltcheck.SaltCheck()
            aaa = 100
            bbb = 110
            mybool = sc_instance._SaltCheck__assert_greater_equal(aaa, bbb)
            self.assertNotEqual(mybool, True)

    def test__assert_greater_equal3(self):
        """test"""
        with patch.dict(
            saltcheck.__salt__,
            {
                "config.get": MagicMock(return_value=True),
                "cp.cache_master": MagicMock(return_value=[True]),
            },
        ):
            sc_instance = saltcheck.SaltCheck()
            aaa = 100
            bbb = 100
            mybool = sc_instance._SaltCheck__assert_greater_equal(aaa, bbb)
            self.assertEqual(mybool, "Pass")

    def test__assert_less1(self):
        """test"""
        with patch.dict(
            saltcheck.__salt__,
            {
                "config.get": MagicMock(return_value=True),
                "cp.cache_master": MagicMock(return_value=[True]),
            },
        ):
            sc_instance = saltcheck.SaltCheck()
            aaa = 99
            bbb = 100
            mybool = sc_instance._SaltCheck__assert_less(aaa, bbb)
            self.assertTrue(mybool, True)

    def test__assert_less2(self):
        """test"""
        with patch.dict(
            saltcheck.__salt__,
            {
                "config.get": MagicMock(return_value=True),
                "cp.cache_master": MagicMock(return_value=[True]),
            },
        ):
            sc_instance = saltcheck.SaltCheck()
            aaa = 110
            bbb = 99
            mybool = sc_instance._SaltCheck__assert_less(aaa, bbb)
            self.assertNotEqual(mybool, True)

    def test__assert_less3(self):
        """test"""
        with patch.dict(
            saltcheck.__salt__,
            {
                "config.get": MagicMock(return_value=True),
                "cp.cache_master": MagicMock(return_value=[True]),
            },
        ):
            sc_instance = saltcheck.SaltCheck()
            aaa = 100
            bbb = 100
            mybool = sc_instance._SaltCheck__assert_less(aaa, bbb)
            self.assertNotEqual(mybool, True)

    def test__assert_less_equal1(self):
        """test"""
        with patch.dict(
            saltcheck.__salt__,
            {
                "config.get": MagicMock(return_value=True),
                "cp.cache_master": MagicMock(return_value=[True]),
            },
        ):
            sc_instance = saltcheck.SaltCheck()
            aaa = 99
            bbb = 100
            mybool = sc_instance._SaltCheck__assert_less_equal(aaa, bbb)
            self.assertTrue(mybool, True)

    def test__assert_less_equal2(self):
        """test"""
        with patch.dict(
            saltcheck.__salt__,
            {
                "config.get": MagicMock(return_value=True),
                "cp.cache_master": MagicMock(return_value=[True]),
            },
        ):
            sc_instance = saltcheck.SaltCheck()
            aaa = 110
            bbb = 99
            mybool = sc_instance._SaltCheck__assert_less_equal(aaa, bbb)
            self.assertNotEqual(mybool, True)

    def test__assert_less_equal3(self):
        """test"""
        with patch.dict(
            saltcheck.__salt__,
            {
                "config.get": MagicMock(return_value=True),
                "cp.cache_master": MagicMock(return_value=[True]),
            },
        ):
            sc_instance = saltcheck.SaltCheck()
            aaa = 100
            bbb = 100
            mybool = sc_instance._SaltCheck__assert_less_equal(aaa, bbb)
            self.assertEqual(mybool, "Pass")

    def test__assert_empty(self):
        """test"""
        with patch.dict(
            saltcheck.__salt__,
            {
                "config.get": MagicMock(return_value=True),
                "cp.cache_master": MagicMock(return_value=[True]),
            },
        ):
            sc_instance = saltcheck.SaltCheck()
            mybool = sc_instance._SaltCheck__assert_empty("")
            self.assertEqual(mybool, "Pass")

    def test__assert_empty_fail(self):
        """test"""
        with patch.dict(
            saltcheck.__salt__,
            {
                "config.get": MagicMock(return_value=True),
                "cp.cache_master": MagicMock(return_value=[True]),
            },
        ):
            sc_instance = saltcheck.SaltCheck()
            mybool = sc_instance._SaltCheck__assert_empty("data")
            self.assertNotEqual(mybool, "Pass")

    def test__assert__not_empty(self):
        """test"""
        with patch.dict(
            saltcheck.__salt__,
            {
                "config.get": MagicMock(return_value=True),
                "cp.cache_master": MagicMock(return_value=[True]),
            },
        ):
            sc_instance = saltcheck.SaltCheck()
            mybool = sc_instance._SaltCheck__assert_not_empty("data")
            self.assertEqual(mybool, "Pass")

    def test__assert__not_empty_fail(self):
        """test"""
        with patch.dict(
            saltcheck.__salt__,
            {
                "config.get": MagicMock(return_value=True),
                "cp.cache_master": MagicMock(return_value=[True]),
            },
        ):
            sc_instance = saltcheck.SaltCheck()
            mybool = sc_instance._SaltCheck__assert_not_empty("")
            self.assertNotEqual(mybool, "Pass")

    @pytest.mark.slow_test
    def test_run_test_1(self):
        """test"""
        with patch.dict(
            saltcheck.__salt__,
            {
                "config.get": MagicMock(return_value=True),
                "sys.list_modules": MagicMock(return_value=["test"]),
                "sys.list_functions": MagicMock(return_value=["test.echo"]),
                "cp.cache_master": MagicMock(return_value=[True]),
            },
        ):
            returned = saltcheck.run_test(
                test={
                    "module_and_function": "test.echo",
                    "assertion": "assertEqual",
                    "expected_return": "This works!",
                    "args": ["This works!"],
                }
            )
            self.assertEqual(returned["status"], "Pass")

    def test_run_test_muliassert(self):
        """test"""
        with patch.dict(
            saltcheck.__salt__,
            {
                "config.get": MagicMock(return_value=True),
                "sys.list_modules": MagicMock(return_value=["test"]),
                "sys.list_functions": MagicMock(return_value=["test.echo"]),
                "cp.cache_master": MagicMock(return_value=[True]),
            },
        ):
            returned = saltcheck.run_test(
                test={
                    "module_and_function": "test.echo",
                    "assertions": [
                        {"assertion": "assertEqual", "expected_return": "This works!"},
                        {"assertion": "assertEqual", "expected_return": "This works!"},
                    ],
                    "args": ["This works!"],
                }
            )
            self.assertEqual(returned["status"], "Pass")

    def test_run_test_muliassert_failure(self):
        """test"""
        with patch.dict(
            saltcheck.__salt__,
            {
                "config.get": MagicMock(return_value=True),
                "sys.list_modules": MagicMock(return_value=["test"]),
                "sys.list_functions": MagicMock(return_value=["test.echo"]),
                "cp.cache_master": MagicMock(return_value=[True]),
            },
        ):
            returned = saltcheck.run_test(
                test={
                    "module_and_function": "test.echo",
                    "assertions": [
                        {"assertion": "assertEqual", "expected_return": "WRONG"},
                        {"assertion": "assertEqual", "expected_return": "This works!"},
                    ],
                    "args": ["This works!"],
                }
            )
            self.assertEqual(returned["status"], "Fail")

    def test_report_highstate_tests(self):
        """test report_highstate_tests"""
        expected_output = {
            "TEST REPORT RESULTS": {
                "States missing tests": ["state1"],
                "Missing Tests": 1,
                "States with tests": ["found"],
            }
        }
        with patch("salt.modules.saltcheck._get_top_states") as mocked_get_top:
            mocked_get_top.return_value = ["state1", "found"]
            with patch("salt.modules.saltcheck.StateTestLoader") as mocked_stl:
                instance = mocked_stl.return_value
                instance.found_states = ["found"]
                returned = saltcheck.report_highstate_tests()
                self.assertEqual(returned, expected_output)

    def test_validation(self):
        """test validation of tests"""
        sc_instance = saltcheck.SaltCheck()

        # Fail on empty test
        test_dict = {}
        expected_return = False
        val_ret = sc_instance._SaltCheck__is_valid_test(test_dict)
        self.assertEqual(val_ret, expected_return)

        # Succeed on standard test
        test_dict = {
            "module_and_function": "test.echo",
            "args": ["hello"],
            "kwargs": {},
            "assertion": "assertEqual",
            "expected_return": "hello",
        }
        expected_return = True
        with patch.dict(
            saltcheck.__salt__,
            {
                "sys.list_modules": MagicMock(return_value=["test"]),
                "sys.list_functions": MagicMock(return_value=["test.echo"]),
            },
        ):
            val_ret = sc_instance._SaltCheck__is_valid_test(test_dict)
            self.assertEqual(val_ret, expected_return)

        # Succeed on standard test with older expected-return syntax
        test_dict = {
            "module_and_function": "test.echo",
            "args": ["hello"],
            "kwargs": {},
            "assertion": "assertEqual",
            "expected-return": "hello",
        }
        expected_return = True
        with patch.dict(
            saltcheck.__salt__,
            {
                "sys.list_modules": MagicMock(return_value=["test"]),
                "sys.list_functions": MagicMock(return_value=["test.echo"]),
            },
        ):
            val_ret = sc_instance._SaltCheck__is_valid_test(test_dict)
            self.assertEqual(val_ret, expected_return)

        # Do not require expected_return for some assertions
        assertions = ["assertEmpty", "assertNotEmpty", "assertTrue", "assertFalse"]
        for assertion in assertions:
            test_dict = {"module_and_function": "test.echo", "args": ["hello"]}
            test_dict["assertion"] = assertion
            expected_return = True
            with patch.dict(
                saltcheck.__salt__,
                {
                    "sys.list_modules": MagicMock(return_value=["test"]),
                    "sys.list_functions": MagicMock(return_value=["test.echo"]),
                },
            ):
                val_ret = sc_instance._SaltCheck__is_valid_test(test_dict)
                self.assertEqual(val_ret, expected_return)

        # Fail on invalid module
        test_dict = {
            "module_and_function": "broken.echo",
            "args": ["hello"],
            "kwargs": {},
            "assertion": "assertEqual",
            "expected_return": "hello",
        }
        expected_return = False
        with patch.dict(
            saltcheck.__salt__,
            {
                "sys.list_modules": MagicMock(return_value=["test"]),
                "sys.list_functions": MagicMock(return_value=["test.echo"]),
            },
        ):
            val_ret = sc_instance._SaltCheck__is_valid_test(test_dict)
            self.assertEqual(val_ret, expected_return)

        # Fail on invalid function
        test_dict = {
            "module_and_function": "test.broken",
            "args": ["hello"],
            "kwargs": {},
            "assertion": "assertEqual",
            "expected_return": "hello",
        }
        expected_return = False
        with patch.dict(
            saltcheck.__salt__,
            {
                "sys.list_modules": MagicMock(return_value=["test"]),
                "sys.list_functions": MagicMock(return_value=["test.echo"]),
            },
        ):
            val_ret = sc_instance._SaltCheck__is_valid_test(test_dict)
            self.assertEqual(val_ret, expected_return)

        # Fail on missing expected_return
        test_dict = {
            "module_and_function": "test.echo",
            "args": ["hello"],
            "kwargs": {},
            "assertion": "assertEqual",
        }
        expected_return = False
        with patch.dict(
            saltcheck.__salt__,
            {
                "sys.list_modules": MagicMock(return_value=["test"]),
                "sys.list_functions": MagicMock(return_value=["test.echo"]),
            },
        ):
            val_ret = sc_instance._SaltCheck__is_valid_test(test_dict)
            self.assertEqual(val_ret, expected_return)

        # Fail on empty expected_return
        test_dict = {
            "module_and_function": "test.echo",
            "args": ["hello"],
            "kwargs": {},
            "assertion": "assertEqual",
            "expected_return": None,
        }
        expected_return = False
        with patch.dict(
            saltcheck.__salt__,
            {
                "sys.list_modules": MagicMock(return_value=["test"]),
                "sys.list_functions": MagicMock(return_value=["test.echo"]),
            },
        ):
            val_ret = sc_instance._SaltCheck__is_valid_test(test_dict)
            self.assertEqual(val_ret, expected_return)

        # Succeed on m_and_f saltcheck.state_apply with only args
        test_dict = {"module_and_function": "saltcheck.state_apply", "args": ["common"]}
        expected_return = True
        with patch.dict(
            saltcheck.__salt__,
            {
                "sys.list_modules": MagicMock(return_value=["saltcheck"]),
                "sys.list_functions": MagicMock(return_value=["saltcheck.state_apply"]),
            },
        ):
            val_ret = sc_instance._SaltCheck__is_valid_test(test_dict)
            self.assertEqual(val_ret, expected_return)

        # Succeed on multiple assertions
        test_dict = {
            "module_and_function": "test.echo",
            "args": ["somearg"],
            "assertions": [
                {
                    "assertion": "assertEqual",
                    "assertion_section": "0:program",
                    "expected_return": "systemd-resolve",
                },
                {
                    "assertion": "assertEqual",
                    "assertion_section": "0:proto",
                    "expected_return": "udp",
                },
            ],
        }
        expected_return = True
        with patch.dict(
            saltcheck.__salt__,
            {
                "sys.list_modules": MagicMock(return_value=["test"]),
                "sys.list_functions": MagicMock(return_value=["test.echo"]),
            },
        ):
            val_ret = sc_instance._SaltCheck__is_valid_test(test_dict)
            self.assertEqual(val_ret, expected_return)

    def test_sls_path_generation(self):
        """test generation of sls paths"""
        with patch.dict(
            saltcheck.__salt__,
            {"config.get": MagicMock(return_value="saltcheck-tests")},
        ):
            testLoader = saltcheck.StateTestLoader()

            state_name = "teststate"
            expected_return = [
                "salt://teststate/saltcheck-tests",
                "salt:///saltcheck-tests",
            ]
            ret = testLoader._generate_sls_path(state_name)
            self.assertEqual(ret, expected_return)

            state_name = "teststate.long.path"
            expected_return = [
                "salt://teststate/long/path/saltcheck-tests",
                "salt://teststate/long/saltcheck-tests",
                "salt://teststate/saltcheck-tests",
            ]
            ret = testLoader._generate_sls_path(state_name)
            self.assertEqual(ret, expected_return)

            state_name = "teststate.really.long.path"
            expected_return = [
                "salt://teststate/really/long/path/saltcheck-tests",
                "salt://teststate/really/long/saltcheck-tests",
                "salt://teststate/saltcheck-tests",
            ]
            ret = testLoader._generate_sls_path(state_name)
            self.assertEqual(ret, expected_return)

    def test_generate_output(self):
        # passing states
        sc_results = {
            "a_state": {
                "test_id1": {"status": "Pass", "duration": 1.987},
                "test_id2": {"status": "Pass", "duration": 1.123},
            }
        }
        expected_output = [
            {
                "a_state": {
                    "test_id1": {"status": "Pass", "duration": 1.987},
                    "test_id2": {"status": "Pass", "duration": 1.123},
                }
            },
            {
                "TEST RESULTS": {
                    "Execution Time": 3.11,
                    "Passed": 2,
                    "Failed": 0,
                    "Skipped": 0,
                    "Missing Tests": 0,
                }
            },
        ]
        ret = saltcheck._generate_out_list(sc_results)
        self.assertEqual(ret, expected_output)

        # Skipped
        sc_results = {
            "a_state": {
                "test_id1": {"status": "Skip", "duration": 1.987},
                "test_id2": {"status": "Pass", "duration": 1.123},
            }
        }
        expected_output = [
            {
                "a_state": {
                    "test_id1": {"status": "Skip", "duration": 1.987},
                    "test_id2": {"status": "Pass", "duration": 1.123},
                }
            },
            {
                "TEST RESULTS": {
                    "Execution Time": 3.11,
                    "Passed": 1,
                    "Failed": 0,
                    "Skipped": 1,
                    "Missing Tests": 0,
                }
            },
        ]
        ret = saltcheck._generate_out_list(sc_results)
        self.assertEqual(ret, expected_output)

        # Failed (does not test setting __context__)
        sc_results = {
            "a_state": {
                "test_id1": {"status": "Failed", "duration": 1.987},
                "test_id2": {"status": "Pass", "duration": 1.123},
            }
        }
        expected_output = [
            {
                "a_state": {
                    "test_id1": {"status": "Failed", "duration": 1.987},
                    "test_id2": {"status": "Pass", "duration": 1.123},
                }
            },
            {
                "TEST RESULTS": {
                    "Execution Time": 3.11,
                    "Passed": 1,
                    "Failed": 1,
                    "Skipped": 0,
                    "Missing Tests": 0,
                }
            },
        ]
        ret = saltcheck._generate_out_list(sc_results)
        self.assertEqual(ret, expected_output)

        # missing states
        sc_results = {
            "a_state": {
                "test_id1": {"status": "Pass", "duration": 1.987},
                "test_id2": {"status": "Pass", "duration": 1.123},
            },
            "b_state": {},
        }
        expected_output = [
            {
                "a_state": {
                    "test_id1": {"status": "Pass", "duration": 1.987},
                    "test_id2": {"status": "Pass", "duration": 1.123},
                }
            },
            {"b_state": {}},
            {
                "TEST RESULTS": {
                    "Execution Time": 3.11,
                    "Passed": 2,
                    "Failed": 0,
                    "Skipped": 0,
                    "Missing Tests": 1,
                }
            },
        ]
        ret = saltcheck._generate_out_list(sc_results)
        self.assertEqual(ret, expected_output)

        # Failed with only_fails
        sc_results = {
            "a_state": {
                "test_id1": {
                    "assertion1": {
                        "status": "Pass",
                        "module.function [args]": 'cmd.run ["ls /etc/salt/master"]',
                        "saltcheck assertion": "IS NOT EMPTY /etc/salt/master",
                    },
                    "assertion2": {
                        "status": "Pass",
                        "module.function [args]": 'cmd.run ["ls /etc/salt/master"]',
                        "saltcheck assertion": "master IN /etc/salt/master",
                    },
                    "status": "Pass",
                    "duration": 1.4383,
                },
                "test_id2": {
                    "assertion1": {"status": "Fail: /etc/salt/master is not empty"},
                    "status": "Fail",
                    "duration": 0.308,
                },
                "test_id3": {
                    "assertion1": {"status": "Fail: /etc/salt/master is not empty"},
                    "status": "Fail",
                    "duration": 0.3073,
                },
            },
            "b_state": {
                "test_id4": {
                    "assertion1": {
                        "status": "Pass",
                        "module.function [args]": 'cmd.run ["ls /etc/salt/master"]',
                        "saltcheck assertion": "IS NOT EMPTY /etc/salt/master",
                    },
                    "assertion2": {
                        "status": "Pass",
                        "module.function [args]": 'cmd.run ["ls /etc/salt/master"]',
                        "saltcheck assertion": "master IN /etc/salt/master",
                    },
                    "status": "Pass",
                    "duration": 0.3057,
                },
                "test_id5": {
                    "assertion1": {"status": "Fail: /etc/salt/master is not empty"},
                    "status": "Fail",
                    "duration": 0.3066,
                },
                "test_id6": {
                    "assertion1": {"status": "Fail: /etc/salt/master is not empty"},
                    "status": "Fail",
                    "duration": 0.3076,
                },
            },
        }
        expected_output = [
            {
                "a_state": {
                    "test_id2": {
                        "assertion1": {"status": "Fail: /etc/salt/master is not empty"},
                        "status": "Fail",
                        "duration": 0.308,
                    },
                    "test_id3": {
                        "assertion1": {"status": "Fail: /etc/salt/master is not empty"},
                        "status": "Fail",
                        "duration": 0.3073,
                    },
                }
            },
            {
                "b_state": {
                    "test_id5": {
                        "assertion1": {"status": "Fail: /etc/salt/master is not empty"},
                        "status": "Fail",
                        "duration": 0.3066,
                    },
                    "test_id6": {
                        "assertion1": {"status": "Fail: /etc/salt/master is not empty"},
                        "status": "Fail",
                        "duration": 0.3076,
                    },
                }
            },
            {
                "TEST RESULTS": {
                    "Execution Time": 2.9735,
                    "Passed": 2,
                    "Failed": 4,
                    "Skipped": 0,
                    "Missing Tests": 0,
                }
            },
        ]
        ret = saltcheck._generate_out_list(sc_results, only_fails=True)
        self.assertEqual(ret, expected_output)



"""
    Taken 1:1 from test cases for salt.modules.config
    This tests the SSH wrapper module.
"""

import fnmatch

import pytest

import salt.client.ssh.wrapper.config as config
from tests.support.mock import patch


@pytest.fixture
def defaults():
    return {
        "test.option.foo": "value of test.option.foo in defaults",
        "test.option.bar": "value of test.option.bar in defaults",
        "test.option.baz": "value of test.option.baz in defaults",
        "test.option": "value of test.option in defaults",
    }


@pytest.fixture
def no_match():
    return "test.option.nope"


@pytest.fixture
def opt_name():
    return "test.option.foo"


@pytest.fixture
def wildcard_opt_name():
    return "test.option.b*"


@pytest.fixture
def configure_loader_modules():
    return {
        config: {
            "__opts__": {
                "test.option.foo": "value of test.option.foo in __opts__",
                "test.option.bar": "value of test.option.bar in __opts__",
                "test.option.baz": "value of test.option.baz in __opts__",
            },
            "__pillar__": {
                "test.option.foo": "value of test.option.foo in __pillar__",
                "test.option.bar": "value of test.option.bar in __pillar__",
                "test.option.baz": "value of test.option.baz in __pillar__",
                "master": {
                    "test.option.foo": "value of test.option.foo in master",
                    "test.option.bar": "value of test.option.bar in master",
                    "test.option.baz": "value of test.option.baz in master",
                },
            },
            "__grains__": {
                "test.option.foo": "value of test.option.foo in __grains__",
                "test.option.bar": "value of test.option.bar in __grains__",
                "test.option.baz": "value of test.option.baz in __grains__",
            },
        }
    }


def _wildcard_match(data, wildcard_opt_name):
    return {x: data[x] for x in fnmatch.filter(data, wildcard_opt_name)}


def test_defaults_only_name(defaults):
    with patch.dict(config.DEFAULTS, defaults):
        opt_name = "test.option"
        opt = config.option(opt_name)
        assert opt == config.DEFAULTS[opt_name]


def test_no_match(defaults, no_match, wildcard_opt_name):
    """
    Make sure that the defa
    """
    with patch.dict(config.DEFAULTS, defaults):
        ret = config.option(no_match)
        assert ret == "", ret

        default = "wat"
        ret = config.option(no_match, default=default)
        assert ret == default, ret

        ret = config.option(no_match, wildcard=True)
        assert ret == {}, ret

        default = {"foo": "bar"}
        ret = config.option(no_match, default=default, wildcard=True)
        assert ret == default, ret

        # Should be no match since wildcard=False
        ret = config.option(wildcard_opt_name)
        assert ret == "", ret


def test_omits(defaults, opt_name, wildcard_opt_name):
    with patch.dict(config.DEFAULTS, defaults):

        # ********** OMIT NOTHING **********

        # Match should be in __opts__ dict
        ret = config.option(opt_name)
        assert ret == config.__opts__[opt_name], ret

        # Wildcard match
        ret = config.option(wildcard_opt_name, wildcard=True)
        assert ret == _wildcard_match(config.__opts__, wildcard_opt_name), ret

        # ********** OMIT __opts__ **********

        # Match should be in __grains__ dict
        ret = config.option(opt_name, omit_opts=True)
        assert ret == config.__grains__[opt_name], ret

        # Wildcard match
        ret = config.option(wildcard_opt_name, omit_opts=True, wildcard=True)
        assert ret == _wildcard_match(config.__grains__, wildcard_opt_name), ret

        # ********** OMIT __opts__, __grains__ **********

        # Match should be in __pillar__ dict
        ret = config.option(opt_name, omit_opts=True, omit_grains=True)
        assert ret == config.__pillar__[opt_name], ret

        # Wildcard match
        ret = config.option(
            wildcard_opt_name, omit_opts=True, omit_grains=True, wildcard=True
        )
        assert ret == _wildcard_match(config.__pillar__, wildcard_opt_name), ret

        # ********** OMIT __opts__, __grains__, __pillar__ **********

        # Match should be in master opts
        ret = config.option(
            opt_name, omit_opts=True, omit_grains=True, omit_pillar=True
        )
        assert ret == config.__pillar__["master"][opt_name], ret

        # Wildcard match
        ret = config.option(
            wildcard_opt_name,
            omit_opts=True,
            omit_grains=True,
            omit_pillar=True,
            wildcard=True,
        )
        assert ret == _wildcard_match(
            config.__pillar__["master"], wildcard_opt_name
        ), ret

        # ********** OMIT ALL THE THINGS **********

        # Match should be in master opts
        ret = config.option(
            opt_name,
            omit_opts=True,
            omit_grains=True,
            omit_pillar=True,
            omit_master=True,
        )
        assert ret == config.DEFAULTS[opt_name], ret

        # Wildcard match
        ret = config.option(
            wildcard_opt_name,
            omit_opts=True,
            omit_grains=True,
            omit_pillar=True,
            omit_master=True,
            wildcard=True,
        )
        assert ret == _wildcard_match(config.DEFAULTS, wildcard_opt_name), ret

        # Match should be in master opts
        ret = config.option(opt_name, omit_all=True)
        assert ret == config.DEFAULTS[opt_name], ret

        # Wildcard match
        ret = config.option(wildcard_opt_name, omit_all=True, wildcard=True)
        assert ret == _wildcard_match(config.DEFAULTS, wildcard_opt_name), ret


# --- Additional tests not found in the execution module tests


@pytest.mark.parametrize("backup", ("", "minion", "master", "both"))
def test_backup_mode(backup):
    res = config.backup_mode(backup)
    assert res == backup or "minion"


@pytest.mark.parametrize(
    "uri,expected",
    (("salt://my/foo.txt", True), ("mysql://foo:bar@foo.bar/baz", False)),
)
def test_valid_fileproto(uri, expected):
    res = config.valid_fileproto(uri)
    assert res is expected


def test_dot_vals():
    extra_master_opt = ("test.option.baah", "value of test.option.baah in master")
    with patch.dict(config.__pillar__, {"master": dict((extra_master_opt,))}):
        res = config.dot_vals("test")
    assert isinstance(res, dict)
    assert res
    for var in ("foo", "bar", "baz"):
        key = f"test.option.{var}"
        assert key in res
        assert res[key] == f"value of test.option.{var} in __opts__"
    assert extra_master_opt[0] in res
    assert res[extra_master_opt[0]] == extra_master_opt[1]



import urllib.parse

import attr
import pytest

import salt.utils.json
import salt.utils.yaml
import tests.support.netapi as netapi
from salt.netapi.rest_tornado import saltnado

pytestmark = [
    pytest.mark.destructive_test,
    pytest.mark.skip_if_not_root,
]


ACCOUNT_USERNAME = "saltdev-syntax"
ACCOUNT_GROUP_NAME = f"{ACCOUNT_USERNAME}-group"


@attr.s(frozen=True, slots=True)
class ExternalAuthConfig:
    eauth = attr.ib()
    pam_key = attr.ib(repr=False)
    pam_config = attr.ib(repr=False)
    expected_perms = attr.ib(repr=False)
    fixture_id = attr.ib(repr=False)
    auto = attr.ib(init=False)
    pam = attr.ib(init=False)

    @auto.default
    def _set_auto(self):
        return {
            "*": ["grains.*"],
            ACCOUNT_USERNAME: ["@wheel"],
            f"{ACCOUNT_GROUP_NAME}%": ["@runner"],
        }

    @pam.default
    def _set_pam(self):
        return {self.pam_key: self.pam_config}


@pytest.fixture(scope="module")
def netapi_account():
    with pytest.helpers.create_account(
        username=ACCOUNT_USERNAME, password="saltdev", group_name=ACCOUNT_GROUP_NAME
    ) as account:
        yield account


def external_auth_ids(value):
    return value.fixture_id


@pytest.fixture(
    params=(
        # By User
        ExternalAuthConfig(
            eauth="pam",
            pam_key=ACCOUNT_USERNAME,
            pam_config=["test.*"],
            expected_perms=["test.*"],
            fixture_id="by-user-pam",
        ),
        ExternalAuthConfig(
            eauth="auto",
            pam_key=ACCOUNT_USERNAME,
            pam_config=["test.*"],
            expected_perms=["@wheel", "grains.*"],
            fixture_id="by-user-auto",
        ),
        # By Group
        ExternalAuthConfig(
            eauth="pam",
            pam_key=f"{ACCOUNT_GROUP_NAME}%",
            pam_config=["grains.*"],
            expected_perms=["grains.*"],
            fixture_id="by-group-pam",
        ),
        ExternalAuthConfig(
            eauth="auto",
            pam_key=f"{ACCOUNT_GROUP_NAME}%",
            pam_config=["@wheel", "grains.*"],
            expected_perms=["@wheel", "grains.*"],
            fixture_id="by-group-auto",
        ),
        # By user, by minion
        ExternalAuthConfig(
            eauth="pam",
            pam_key=ACCOUNT_USERNAME,
            pam_config=[{"G@id:master2": ["@jobs"]}, {"G@id:master1": ["@jobs"]}],
            expected_perms=[{"G@id:master1": ["@jobs"]}, {"G@id:master2": ["@jobs"]}],
            fixture_id="by-user-by-minion-pam",
        ),
        ExternalAuthConfig(
            eauth="auto",
            pam_key=ACCOUNT_USERNAME,
            pam_config=[{"G@id:master2": ["@jobs"]}, {"G@id:master1": ["@jobs"]}],
            expected_perms=["@wheel", "grains.*"],
            fixture_id="by-user-by-minion-auto",
        ),
        # By user, by wheel
        ExternalAuthConfig(
            eauth="pam",
            pam_key=ACCOUNT_USERNAME,
            pam_config=["@wheel"],
            expected_perms=["@wheel"],
            fixture_id="by-user-by-@wheel-pam",
        ),
        ExternalAuthConfig(
            eauth="auto",
            pam_key=ACCOUNT_USERNAME,
            pam_config=["@wheel"],
            expected_perms=["@wheel", "grains.*"],
            fixture_id="by-user-by-@wheel-auto",
        ),
        # By user, by runner
        ExternalAuthConfig(
            eauth="pam",
            pam_key=ACCOUNT_USERNAME,
            pam_config=["@runner"],
            expected_perms=["@runner"],
            fixture_id="by-user-by-@runner-pam",
        ),
        ExternalAuthConfig(
            eauth="auto",
            pam_key=ACCOUNT_USERNAME,
            pam_config=["@runner"],
            expected_perms=["@wheel", "grains.*"],
            fixture_id="by-user-by-@runner-auto",
        ),
        # By user, by jobs
        ExternalAuthConfig(
            eauth="pam",
            pam_key=ACCOUNT_USERNAME,
            pam_config=["@jobs"],
            expected_perms=["@jobs"],
            fixture_id="by-user-by-@jobs-pam",
        ),
        ExternalAuthConfig(
            eauth="auto",
            pam_key=ACCOUNT_USERNAME,
            pam_config=["@jobs"],
            expected_perms=["@wheel", "grains.*"],
            fixture_id="by-user-by-@jobs-auto",
        ),
        # By group, by wheel
        ExternalAuthConfig(
            eauth="pam",
            pam_key=f"{ACCOUNT_GROUP_NAME}%",
            pam_config=["@wheel"],
            expected_perms=["@wheel"],
            fixture_id="by-group-by-@wheel-pam",
        ),
        ExternalAuthConfig(
            eauth="auto",
            pam_key=f"{ACCOUNT_GROUP_NAME}%",
            pam_config=["@wheel"],
            expected_perms=["@wheel", "grains.*"],
            fixture_id="by-group-by-@wheel-auto",
        ),
        # By group, by runner
        ExternalAuthConfig(
            eauth="pam",
            pam_key=f"{ACCOUNT_GROUP_NAME}%",
            pam_config=["@runner"],
            expected_perms=["@runner"],
            fixture_id="by-group-by-@runner-pam",
        ),
        ExternalAuthConfig(
            eauth="auto",
            pam_key=f"{ACCOUNT_GROUP_NAME}%",
            pam_config=["@runner"],
            expected_perms=["@wheel", "grains.*"],
            fixture_id="by-group-by-@runner-auto",
        ),
        # By group, by jobs
        ExternalAuthConfig(
            eauth="pam",
            pam_key=f"{ACCOUNT_GROUP_NAME}%",
            pam_config=["@jobs"],
            expected_perms=["@jobs"],
            fixture_id="by-group-by-@jobs-pam",
        ),
        ExternalAuthConfig(
            eauth="auto",
            pam_key=f"{ACCOUNT_GROUP_NAME}%",
            pam_config=["@jobs"],
            expected_perms=["@wheel", "grains.*"],
            fixture_id="by-group-by-@jobs-auto",
        ),
        # By user, by wheel/runner/jobs module
        ExternalAuthConfig(
            eauth="pam",
            pam_key=ACCOUNT_USERNAME,
            pam_config=[{"@runner": ["active"]}],
            expected_perms=[{"@runner": ["active"]}],
            fixture_id="by-user-by-@wheel/@runner/@jobs-module-pam",
        ),
        ExternalAuthConfig(
            eauth="auto",
            pam_key=ACCOUNT_USERNAME,
            pam_config=[{"@runner": ["active"]}],
            expected_perms=["@wheel", "grains.*"],
            fixture_id="by-user-by-@wheel/@runner/@jobs-module-auto",
        ),
        # By user, module, args & kwargs
        ExternalAuthConfig(
            eauth="pam",
            pam_key=ACCOUNT_USERNAME,
            pam_config=[
                {
                    "*": [
                        {
                            "my_mod.*": {
                                "args": ["a1.*", ".*", "a3.*"],
                                "kwargs": {"kwa": "kwa.*", "kwb": "kwb"},
                            }
                        }
                    ]
                }
            ],
            expected_perms=[
                {
                    "*": [
                        {
                            "my_mod.*": {
                                "args": ["a1.*", ".*", "a3.*"],
                                "kwargs": {"kwa": "kwa.*", "kwb": "kwb"},
                            }
                        }
                    ]
                }
            ],
            fixture_id="by-user-by-module-args-kwargs-pam",
        ),
        ExternalAuthConfig(
            eauth="auto",
            pam_key=ACCOUNT_USERNAME,
            pam_config=[
                {
                    "*": [
                        {
                            "my_mod.*": {
                                "args": ["a1.*", ".*", "a3.*"],
                                "kwargs": {"kwa": "kwa.*", "kwb": "kwb"},
                            }
                        }
                    ]
                }
            ],
            expected_perms=["@wheel", "grains.*"],
            fixture_id="by-user-by-module-args-kwargs-auto",
        ),
    ),
    ids=external_auth_ids,
)
def external_auth(request):
    return request.param


@pytest.fixture
def auth_creds(external_auth, netapi_account):
    return {
        "username": netapi_account.username,
        "password": netapi_account.password,
        "eauth": external_auth.eauth,
    }


@pytest.fixture
def client_config(client_config, external_auth):
    client_config["external_auth"] = {
        "auto": external_auth.auto,
        "pam": external_auth.pam,
    }
    return client_config


@pytest.fixture
def http_server(io_loop, app, netapi_port, content_type_map):
    with netapi.TestsTornadoHttpServer(
        io_loop=io_loop,
        app=app,
        port=netapi_port,
        client_headers={"Content-Type": content_type_map["form"]},
    ) as server:
        yield server


@pytest.fixture
def app_urls():
    return [
        ("/login", saltnado.SaltAuthHandler),
    ]


async def test_perms(http_client, auth_creds, external_auth):
    response = await http_client.fetch(
        "/login",
        method="POST",
        body=urllib.parse.urlencode(auth_creds),
    )
    assert response.code == 200
    response_obj = salt.utils.json.loads(response.body)["return"][0]
    perms = response_obj["perms"]
    assert perms == external_auth.expected_perms



"""
Integration tests for mac_timezone

If using parallels, make sure Time sync is turned off. Otherwise, parallels will
keep changing your date/time settings while the tests are running. To turn off
Time sync do the following:
    - Go to actions -> configure
    - Select options at the top and 'More Options' on the left
    - Set time to 'Do not sync'
"""

import datetime

import pytest

from salt.exceptions import SaltInvocationError

pytestmark = [
    pytest.mark.skip_if_binaries_missing("systemsetup"),
    pytest.mark.slow_test,
    pytest.mark.destructive_test,
    pytest.mark.skip_if_not_root,
    pytest.mark.skip_unless_on_darwin,
]


@pytest.fixture(scope="module")
def timezone(modules):
    return modules.timezone


@pytest.fixture
def _reset_time_server(timezone):
    ret = timezone.get_time_server()
    try:
        yield
    finally:
        if timezone.get_time_server() != ret:
            timezone.set_time_server(ret)


@pytest.fixture
def _reset_using_network_time(timezone):
    ret = timezone.get_using_network_time()
    try:
        timezone.set_using_network_time(False)
        yield ret
    finally:
        timezone.set_using_network_time(ret)


@pytest.fixture
def _reset_time(timezone, _reset_using_network_time):
    ret = timezone.get_time()
    try:
        yield
    finally:
        if not _reset_using_network_time:
            timezone.set_time(ret)


@pytest.fixture
def _reset_date(timezone, _reset_using_network_time):
    ret = timezone.get_date()
    try:
        yield
    finally:
        if not _reset_using_network_time:
            timezone.set_date(ret)


@pytest.fixture
def _reset_zone(timezone):
    ret = timezone.get_zone()
    try:
        timezone.set_zone("America/Denver")
        yield
    finally:
        timezone.set_zone(ret)


@pytest.mark.usefixtures("_reset_date")
def test_get_set_date(timezone):
    """
    Test timezone.get_date
    Test timezone.set_date
    """
    # Correct Functionality
    ret = timezone.set_date("2/20/2011")
    assert ret
    ret = timezone.get_date()
    assert ret == "2/20/2011"

    # Test bad date format
    with pytest.raises(SaltInvocationError) as exc:
        ret = timezone.set_date("13/12/2014")
        assert (
            "ERROR executing 'timezone.set_date': Invalid Date/Time Format: 13/12/2014"
            in str(exc.value)
        )


@pytest.mark.slow_test
def test_get_time(timezone):
    """
    Test timezone.get_time
    """
    text_time = timezone.get_time()
    assert text_time != "Invalid Timestamp"
    obj_date = datetime.datetime.strptime(text_time, "%H:%M:%S")
    assert isinstance(obj_date, datetime.date)


@pytest.mark.usefixtures("_reset_time")
def test_set_time(timezone):
    """
    Test timezone.set_time
    """
    # Correct Functionality
    ret = timezone.set_time("3:14")
    assert ret

    # Test bad time format
    with pytest.raises(SaltInvocationError) as exc:
        ret = timezone.set_time("3:71")
        assert (
            "ERROR executing 'timezone.set_time': Invalid Date/Time Format: 3:71"
            in str(exc.value)
        )


@pytest.mark.usefixtures("_reset_zone")
def test_get_set_zone(timezone):
    """
    Test timezone.get_zone
    Test timezone.set_zone
    """
    # Correct Functionality
    ret = timezone.set_zone("Pacific/Wake")
    assert ret

    ret = timezone.get_zone()
    assert ret == "Pacific/Wake"

    # Test bad time zone
    with pytest.raises(SaltInvocationError) as exc:
        ret = timezone.set_zone("spongebob")
        assert (
            "ERROR executing 'timezone.set_zone': Invalid Timezone: spongebob"
            in str(exc.value)
        )


@pytest.mark.usefixtures("_reset_zone")
def test_get_offset(timezone):
    """
    Test timezone.get_offset
    """
    pytz = pytest.importorskip("pytz")
    now = datetime.datetime.now(tz=pytz.UTC)

    ret = timezone.set_zone("Pacific/Wake")
    assert ret
    ret = timezone.get_offset()
    assert isinstance(ret, str)
    assert ret == "+1200"

    ret = timezone.set_zone("America/Los_Angeles")
    assert ret
    ret = timezone.get_offset()
    assert isinstance(ret, str)

    if now.astimezone(pytz.timezone("America/Los_Angeles")).dst():
        assert ret == "-0700"
    else:
        assert ret == "-0800"


@pytest.mark.usefixtures("_reset_zone")
def test_get_set_zonecode(timezone):
    """
    Test timezone.get_zonecode
    Test timezone.set_zonecode
    """
    ret = timezone.set_zone("America/Los_Angeles")
    assert ret
    ret = timezone.get_zone()
    assert isinstance(ret, str)
    assert ret == "America/Los_Angeles"

    ret = timezone.set_zone("Pacific/Wake")
    assert ret
    ret = timezone.get_zone()
    assert isinstance(ret, str)
    assert ret == "Pacific/Wake"


@pytest.mark.slow_test
def test_list_zones(timezone):
    """
    Test timezone.list_zones
    """
    zones = timezone.list_zones()
    assert isinstance(zones, list)
    assert "America/Denver" in zones
    assert "America/Los_Angeles" in zones


@pytest.mark.usefixtures("_reset_zone")
def test_zone_compare(timezone):
    """
    Test timezone.zone_compare
    """
    ret = timezone.zone_compare("America/Denver")
    assert ret
    ret = timezone.zone_compare("Pacific/Wake")
    assert not ret


@pytest.mark.usefixtures("_reset_using_network_time")
def test_get_set_using_network_time(timezone):
    """
    Test timezone.get_using_network_time
    Test timezone.set_using_network_time
    """
    ret = timezone.set_using_network_time(True)
    assert ret

    ret = timezone.get_using_network_time()
    assert ret

    ret = timezone.set_using_network_time(False)
    assert ret

    ret = timezone.get_using_network_time()
    assert not ret


@pytest.mark.usefixtures("_reset_time_server")
def test_get_set_time_server(timezone):
    """
    Test timezone.get_time_server
    Test timezone.set_time_server
    """
    ret = timezone.set_time_server("spongebob.com")
    assert ret

    ret = timezone.get_time_server()
    assert ret == "spongebob.com"



import ctypes
import multiprocessing

import pytest
from saltfactories.utils import random_string

import salt.crypt
import salt.master
import salt.utils.stringutils
from tests.conftest import FIPS_TESTRUN


@pytest.fixture(autouse=True)
def _prepare_aes():
    old_aes = salt.master.SMaster.secrets.get("aes")
    try:
        salt.master.SMaster.secrets["aes"] = {
            "secret": multiprocessing.Array(
                ctypes.c_char,
                salt.utils.stringutils.to_bytes(
                    salt.crypt.Crypticle.generate_key_string()
                ),
            ),
            "reload": salt.crypt.Crypticle.generate_key_string,
        }
    finally:
        if old_aes:
            salt.master.SMaster.secrets["aes"] = old_aes


def transport_ids(value):
    return f"Transport({value})"


@pytest.fixture(params=("zeromq", "tcp"), ids=transport_ids)
def transport(request):
    return request.param


@pytest.fixture
def salt_master(salt_factories, transport):
    config_defaults = {
        "transport": transport,
        "auto_accept": True,
        "sign_pub_messages": False,
        "fips_mode": FIPS_TESTRUN,
        "publish_signing_algorithm": (
            "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1"
        ),
    }
    factory = salt_factories.salt_master_daemon(
        random_string(f"server-{transport}-master-"),
        defaults=config_defaults,
    )
    return factory


@pytest.fixture
def salt_minion(salt_master, transport):
    config_defaults = {
        "transport": transport,
        "master_ip": "127.0.0.1",
        "master_port": salt_master.config["ret_port"],
        "auth_timeout": 5,
        "auth_tries": 1,
        "master_uri": f"tcp://127.0.0.1:{salt_master.config['ret_port']}",
        "fips_mode": FIPS_TESTRUN,
        "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
        "signing_algorithm": "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1",
    }
    factory = salt_master.salt_minion_daemon(
        random_string("server-{transport}-minion-"),
        defaults=config_defaults,
    )
    return factory



"""
    :codeauthor: Ryan Lewis (ryansname@gmail.com)

    pytest.unit.modules.portage_flags
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

import pytest

import salt.modules.portage_config as portage_config
import salt.utils.files
from tests.support.mock import patch

pytest.importorskip("portage", reason="System is not gentoo/funtoo.")


def setup_loader_modules():
    return {}


def test_get_config_file_wildcards():
    pairs = [
        ("*/*::repo", "/etc/portage/package.mask/repo"),
        ("*/pkg::repo", "/etc/portage/package.mask/pkg"),
        ("cat/*", "/etc/portage/package.mask/cat_"),
        ("cat/pkg", "/etc/portage/package.mask/cat/pkg"),
        ("cat/pkg::repo", "/etc/portage/package.mask/cat/pkg"),
    ]

    for atom, expected in pairs:
        assert portage_config._get_config_file("mask", atom) == expected


def test_enforce_nice_config(tmp_path):
    atoms = [
        ("*/*::repo", "repo"),
        ("*/pkg1::repo", "pkg1"),
        ("cat/*", "cat_"),
        ("cat/pkg2", "cat/pkg2"),
        ("cat/pkg3::repo", "cat/pkg3"),
        ("<cat/pkg4-0.0.0.0", "cat/pkg4"),
        (">cat/pkg5-0.0.0.0:0", "cat/pkg5"),
        (">cat/pkg6-0.0.0.0:0::repo", "cat/pkg6"),
        ("<=cat/pkg7-0.0.0.0", "cat/pkg7"),
        ("=cat/pkg8-0.0.0.0", "cat/pkg8"),
    ]

    supported = [
        ("accept_keywords", ["~amd64"]),
        ("env", ["glibc.conf"]),
        ("license", ["LICENCE1", "LICENCE2"]),
        ("mask", [""]),
        ("properties", ["* -interactive"]),
        ("unmask", [""]),
        ("use", ["apple", "-banana", "ananas", "orange"]),
    ]

    base_path = str(tmp_path / "/package.{0}")

    def make_line(atom, addition):
        return atom + (" " + addition if addition != "" else "") + "\n"

    for typ, additions in supported:
        path = base_path.format(typ)
        with salt.utils.files.fopen(path, "a") as fh:
            for atom, _ in atoms:
                for addition in additions:
                    line = make_line(atom, addition)
                    fh.write("# comment for: " + line)
                    fh.write(line)

    with patch.object(portage_config, "BASE_PATH", base_path):
        with patch.object(
            portage_config, "_merge_flags", lambda l1, l2, _: list(set(l1 + l2))
        ):
            portage_config.enforce_nice_config()

    for typ, additions in supported:
        for atom, file_name in atoms:
            with salt.utils.files.fopen(
                base_path.format(typ) + "/" + file_name, "r"
            ) as fh:
                for line in fh:
                    for atom in line:
                        assert atom not in line
                    for addition in additions:
                        assert addition not in line



import pytest

import salt.client.ssh.wrapper as wrap


def test_parse_ret_permission_denied_scp():
    """
    Ensure that permission denied errors are raised when scp fails to copy
    a file to a target because of an authentication failure.
    """
    stdout = "\rroot@192.168.1.187's password: \n\rroot@192.168.1.187's password: \n\rroot@192.168.1.187's password: \n"
    stderr = "Permission denied, please try again.\nPermission denied, please try again.\nroot@192.168.1.187: Permission denied (publickey,gssapi-keyex,gssapi-with-micimport pudb; pu.dbassword).\nscp: Connection closed\n"
    retcode = 255

    with pytest.raises(wrap.SSHPermissionDeniedError) as exc:
        wrap.parse_ret(stdout, stderr, retcode)
    ret = exc.value.to_ret()
    assert "_error" in ret
    assert ret["_error"] == "Permission denied"
    assert "stdout" in ret
    assert "stderr" in ret
    assert "Permission denied (publickey" in ret["stderr"]
    assert "retcode" in ret
    assert ret["retcode"] == 255


def test_parse_ret_permission_denied_because_of_permissions():
    """
    Ensure that permission denied errors are NOT raised when scp fails
    to copy a file to a target due to missing permissions of the user account.
    The PermissionDeniedError should be exclusive to authentication failures and
    not apply to authorization ones.
    """
    stdout = ""
    stderr = 'scp: dest open "/tmp/preflight.sh": Permission denied\nscp: failed to upload file /etc/salt/preflight.sh to /tmp/preflight.sh\n'
    retcode = 1

    try:
        wrap.parse_ret(stdout, stderr, retcode)
    except wrap.SSHPermissionDeniedError:
        pytest.fail("This should not have resulted in an SSHPermissionDeniedError")
    except wrap.SSHCommandExecutionError:
        pass



"""
Generate marathon proxy minion grains.

.. versionadded:: 2015.8.2

"""

import salt.utils.http
import salt.utils.platform

__proxyenabled__ = ["marathon"]
__virtualname__ = "marathon"


def __virtual__():
    if (
        salt.utils.platform.is_proxy()
        and "proxy" in __opts__
        and __opts__["proxy"].get("proxytype") == "marathon"
    ):
        return __virtualname__
    return False


def kernel():
    return {"kernel": "marathon"}


def os():
    return {"os": "marathon"}


def os_family():
    return {"os_family": "marathon"}


def os_data():
    return {"os_data": "marathon"}


def marathon():
    response = salt.utils.http.query(
        "{}/v2/info".format(__opts__["proxy"].get("base_url", "http://locahost:8080")),
        decode_type="json",
        decode=True,
    )
    if not response or "dict" not in response:
        return {"marathon": None}
    return {"marathon": response["dict"]}



r"""
The Linode Cloud Module
=======================

The Linode cloud module is used to interact with the Linode Cloud.

Provider
--------

The following provider parameters are supported:

- **apikey**: (required) The key to use to authenticate with the Linode API.
- **password**: (required) The default password to set on new VMs. Must be 8 characters with at least one lowercase, uppercase, and numeric.
- **poll_interval**: (optional) The rate of time in milliseconds to poll the Linode API for changes. Defaults to ``500``.
- **ratelimit_sleep**: (optional) The time in seconds to wait before retrying after a ratelimit has been enforced. Defaults to ``0``.

.. note::

    APIv3 usage has been removed in favor of APIv4. To move to APIv4 now,
    See the full migration guide
    here https://docs.saltproject.io/en/latest/topics/cloud/linode.html#migrating-to-apiv4.

Set up the provider configuration at ``/etc/salt/cloud.providers`` or ``/etc/salt/cloud.providers.d/linode.conf``:

.. code-block:: yaml

    my-linode-provider:
        driver: linode
        apikey: f4ZsmwtB1c7f85Jdu43RgXVDFlNjuJaeIYV8QMftTqKScEB2vSosFSr...
        password: F00barbazverylongp@ssword

Profile
-------

The following profile parameters are supported:

- **size**: (required) The size of the VM. This should be a Linode instance type ID (i.e. ``g6-standard-2``). Run ``salt-cloud -f avail_sizes my-linode-provider`` for options.
- **location**: (required) The location of the VM. This should be a Linode region (e.g. ``us-east``). Run ``salt-cloud -f avail_locations my-linode-provider`` for options.
- **image**: (required) The image to deploy the boot disk from. This should be an image ID (e.g. ``linode/ubuntu22.04``); official images start with ``linode/``. Run ``salt-cloud -f avail_images my-linode-provider`` for more options.
- **password**: (\*required) The default password for the VM. Must be provided at the profile or provider level.
- **assign_private_ip**: (optional) Whether or not to assign a private IP to the VM. Defaults to ``False``.
- **backups_enabled**: (optional) Whether or not to enable the backup for this VM. Backup can be configured in your Linode account Defaults to ``False``.
- **ssh_interface**: (optional) The interface with which to connect over SSH. Valid options are ``private_ips`` or ``public_ips``. Defaults to ``public_ips``.
- **ssh_pubkey**: (optional) The public key to authorize for SSH with the VM.
- **swap**: (optional) The amount of disk space to allocate for the swap partition. Defaults to ``256``.
- **clonefrom**: (optional) The name of the Linode to clone from.

Set up a profile configuration in ``/etc/salt/cloud.profiles.d/``:

.. code-block:: yaml

    my-linode-profile:
        # a minimal configuration
        provider: my-linode-provider
        size: g6-standard-1
        image: linode/ubuntu22.04
        location: us-east

    my-linode-profile-advanced:
        # an advanced configuration
        provider: my-linode-provider
        size: g6-standard-3
        image: linode/ubuntu22.04
        location: eu-west
        password: bogus123X
        assign_private_ip: true
        ssh_interface: private_ips
        ssh_pubkey: ssh-rsa AAAAB3NzaC1yc2EAAAADAQAB...
        swap_size: 512

Migrating to APIv4
------------------

You will need to generate a new token for your account. See https://www.linode.com/docs/products/tools/api/get-started/#create-an-api-token

There are a few changes to note:
- There has been a general move from label references to ID references. The profile configuration parameters ``location``, ``size``, and ``image`` have moved from being label based references to IDs. See the profile section for more information. In addition to these inputs being changed, ``avail_sizes``, ``avail_locations``, and ``avail_images`` now output options sorted by ID instead of label.
- The ``disk_size`` profile configuration parameter has been deprecated and will not be taken into account when creating new VMs while targeting APIv4.

:maintainer: Linode Developer Tools and Experience Team <dev-dx@linode.com>
:depends: requests
"""

import datetime
import json
import logging
import pprint
import re
import time
from abc import ABC, abstractmethod
from pathlib import Path

import salt.config as config
from salt._compat import ipaddress
from salt.exceptions import SaltCloudException, SaltCloudNotFound, SaltCloudSystemExit

try:
    import requests

    HAS_REQUESTS = True
except ImportError:
    HAS_REQUESTS = False

# Get logging started
log = logging.getLogger(__name__)

# The epoch of the last time a query was made
LASTCALL = int(time.mktime(datetime.datetime.now().timetuple()))

__virtualname__ = "linode"


# Only load in this module if the Linode configurations are in place
def __virtual__():
    """
    Check for Linode configs.
    """
    if get_configured_provider() is False:
        return False

    if _get_dependencies() is False:
        return False

    return __virtualname__


def _get_active_provider_name():
    try:
        return __active_provider_name__.value()
    except AttributeError:
        return __active_provider_name__


def _get_backup_enabled(vm_):
    """
    Return True if a backup is set to enabled
    """
    return config.get_cloud_config_value(
        "backups_enabled",
        vm_,
        __opts__,
        default=False,
    )


def get_configured_provider():
    """
    Return the first configured instance.
    """
    return config.is_provider_configured(
        __opts__,
        _get_active_provider_name() or __virtualname__,
        ("apikey", "password"),
    )


def _get_dependencies():
    """
    Warn if dependencies aren't met.
    """
    deps = {"requests": HAS_REQUESTS}
    return config.check_driver_dependencies(__virtualname__, deps)


def _get_api_key():
    """
    Returned the configured Linode API key.
    """
    val = config.get_cloud_config_value(
        "api_key",
        get_configured_provider(),
        __opts__,
        search_global=False,
        default=config.get_cloud_config_value(
            "apikey", get_configured_provider(), __opts__, search_global=False
        ),
    )
    return val


def _get_ratelimit_sleep():
    """
    Return the configured time to wait before retrying after a ratelimit has been enforced.
    """
    return config.get_cloud_config_value(
        "ratelimit_sleep",
        get_configured_provider(),
        __opts__,
        search_global=False,
        default=0,
    )


def _get_poll_interval():
    """
    Return the configured interval in milliseconds to poll the Linode API for changes at.
    """
    return config.get_cloud_config_value(
        "poll_interval",
        get_configured_provider(),
        __opts__,
        search_global=False,
        default=500,
    )


def _get_password(vm_):
    r"""
    Return the password to use for a VM.

    vm\_
        The configuration to obtain the password from.
    """
    return config.get_cloud_config_value(
        "password",
        vm_,
        __opts__,
        default=config.get_cloud_config_value(
            "passwd", vm_, __opts__, search_global=False
        ),
        search_global=False,
    )


def _get_private_ip(vm_):
    """
    Return True if a private ip address is requested
    """
    return config.get_cloud_config_value(
        "assign_private_ip", vm_, __opts__, default=False
    )


def _get_ssh_key_files(vm_):
    """
    Return the configured file paths of the SSH keys.
    """
    return config.get_cloud_config_value(
        "ssh_key_files", vm_, __opts__, search_global=False, default=[]
    )


def _get_ssh_key(vm_):
    r"""
    Return the SSH pubkey.

    vm\_
        The configuration to obtain the public key from.
    """
    return config.get_cloud_config_value(
        "ssh_pubkey", vm_, __opts__, search_global=False
    )


def _get_swap_size(vm_):
    r"""
    Returns the amount of swap space to be used in MB.

    vm\_
        The VM profile to obtain the swap size from.
    """
    return config.get_cloud_config_value("swap", vm_, __opts__, default=256)


def _get_ssh_keys(vm_):
    """
    Return all SSH keys from ``ssh_pubkey`` and ``ssh_key_files``.
    """
    ssh_keys = set()

    raw_pub_key = _get_ssh_key(vm_)
    if raw_pub_key is not None:
        ssh_keys.add(raw_pub_key)

    key_files = _get_ssh_key_files(vm_)
    for file in map(lambda file: Path(file).resolve(), key_files):
        if not (file.exists() or file.is_file()):
            raise SaltCloudSystemExit(f"Invalid SSH key file: {str(file)}")
        ssh_keys.add(file.read_text())

    return list(ssh_keys)


def _get_ssh_interface(vm_):
    """
    Return the ssh_interface type to connect to. Either 'public_ips' (default)
    or 'private_ips'.
    """
    return config.get_cloud_config_value(
        "ssh_interface", vm_, __opts__, default="public_ips", search_global=False
    )


def _validate_name(name):
    """
    Checks if the provided name fits Linode's labeling parameters.

    .. versionadded:: 2015.5.6

    name
        The VM name to validate
    """
    name = str(name)
    name_length = len(name)
    regex = re.compile(r"^[a-zA-Z0-9][A-Za-z0-9_-]*[a-zA-Z0-9]$")

    if name_length < 3 or name_length > 48:
        ret = False
    elif not re.match(regex, name):
        ret = False
    else:
        ret = True

    if ret is False:
        log.warning(
            "A Linode label may only contain ASCII letters or numbers, dashes, and "
            "underscores, must begin and end with letters or numbers, and be at least "
            "three characters in length."
        )

    return ret


class LinodeAPI(ABC):
    @abstractmethod
    def avail_images(self):
        """avail_images implementation"""

    @abstractmethod
    def avail_locations(self):
        """avail_locations implementation"""

    @abstractmethod
    def avail_sizes(self):
        """avail_sizes implementation"""

    @abstractmethod
    def boot(self, name=None, kwargs=None):
        """boot implementation"""

    @abstractmethod
    def clone(self, kwargs=None):
        """clone implementation"""

    @abstractmethod
    def create_config(self, kwargs=None):
        """create_config implementation"""

    @abstractmethod
    def create(self, vm_):
        """create implementation"""

    @abstractmethod
    def destroy(self, name):
        """destroy implementation"""

    @abstractmethod
    def get_config_id(self, kwargs=None):
        """get_config_id implementation"""

    @abstractmethod
    def list_nodes(self):
        """list_nodes implementation"""

    @abstractmethod
    def list_nodes_full(self):
        """list_nodes_full implementation"""

    @abstractmethod
    def list_nodes_min(self):
        """list_nodes_min implementation"""

    @abstractmethod
    def reboot(self, name):
        """reboot implementation"""

    @abstractmethod
    def show_instance(self, name):
        """show_instance implementation"""

    @abstractmethod
    def show_pricing(self, kwargs=None):
        """show_pricing implementation"""

    @abstractmethod
    def start(self, name):
        """start implementation"""

    @abstractmethod
    def stop(self, name):
        """stop implementation"""

    @abstractmethod
    def _get_linode_by_name(self, name):
        """_get_linode_by_name implementation"""

    @abstractmethod
    def _get_linode_by_id(self, linode_id):
        """_get_linode_by_id implementation"""

    def get_linode(self, kwargs=None):
        name = kwargs.get("name", None)
        linode_id = kwargs.get("linode_id", None)

        if linode_id is not None:
            return self._get_linode_by_id(linode_id)
        elif name is not None:
            return self._get_linode_by_name(name)

        raise SaltCloudSystemExit(
            "The get_linode function requires either a 'name' or a 'linode_id'."
        )

    def list_nodes_select(self, call):
        return __utils__["cloud.list_nodes_select"](
            self.list_nodes_full(),
            __opts__["query.selection"],
            call,
        )


class LinodeAPIv4(LinodeAPI):
    @classmethod
    def get_api_instance(cls):
        if not hasattr(cls, "api_instance"):
            cls.api_instance = cls()
        return cls.api_instance

    def _query(self, path, method="GET", data=None, headers=None):
        """
        Make a call to the Linode API.
        """
        api_key = _get_api_key()
        ratelimit_sleep = _get_ratelimit_sleep()

        if headers is None:
            headers = {}
        headers["Authorization"] = f"Bearer {api_key}"
        headers["Content-Type"] = "application/json"
        headers["User-Agent"] = "salt-cloud-linode"

        url = f"https://api.linode.com/v4{path}"

        decode = method != "DELETE"
        result = None

        log.debug("Linode API request: %s %s", method, url)

        if data is not None:
            log.trace("Linode API request body: %s", data)

        attempt = 0
        while True:
            try:
                result = requests.request(
                    method, url, json=data, headers=headers, timeout=120
                )

                log.debug("Linode API response status code: %d", result.status_code)
                log.trace("Linode API response body: %s", result.text)
                result.raise_for_status()
                break
            except requests.exceptions.HTTPError as exc:
                err_response = exc.response
                err_data = self._get_response_json(err_response)
                status_code = err_response.status_code

                if status_code == 429:
                    log.debug(
                        "received rate limit; retrying in %d seconds", ratelimit_sleep
                    )
                    time.sleep(ratelimit_sleep)
                    continue

                if err_data is not None:
                    # Build an error from the response JSON
                    if "error" in err_data:
                        raise SaltCloudSystemExit(
                            "Linode API reported error: {}".format(err_data["error"])
                        )
                    elif "errors" in err_data:
                        api_errors = err_data["errors"]

                        # Build Salt exception
                        errors = []
                        for error in err_data["errors"]:
                            if "field" in error:
                                errors.append(
                                    "field '{}': {}".format(
                                        error.get("field"), error.get("reason")
                                    )
                                )
                            else:
                                errors.append(error.get("reason"))

                        raise SaltCloudSystemExit(
                            "Linode API reported error(s): {}".format(", ".join(errors))
                        )

                # If the response is not valid JSON or the error was not included, propagate the
                # human readable status representation.
                raise SaltCloudSystemExit(
                    f"Linode API error occurred: {err_response.reason}"
                )
        if decode:
            return self._get_response_json(result)

        return result

    def avail_images(self):
        response = self._query(path="/images")
        ret = {}
        for image in response["data"]:
            ret[image["id"]] = image
        return ret

    def avail_locations(self):
        response = self._query(path="/regions")
        ret = {}
        for region in response["data"]:
            ret[region["id"]] = region
        return ret

    def avail_sizes(self):
        response = self._query(path="/linode/types")
        ret = {}
        for instance_type in response["data"]:
            ret[instance_type["id"]] = instance_type
        return ret

    def set_backup_schedule(self, label, linode_id, day, window, auto_enable=False):
        instance = self.get_linode(kwargs={"linode_id": linode_id, "name": label})
        linode_id = instance.get("id", None)

        if auto_enable:
            backups = instance.get("backups")
            if backups and not backups.get("enabled"):
                self._query(
                    f"/linode/instances/{linode_id}/backups/enable",
                    method="POST",
                )

        self._query(
            f"/linode/instances/{linode_id}",
            method="PUT",
            data={"backups": {"schedule": {"day": day, "window": window}}},
        )

    def boot(self, name=None, kwargs=None):
        instance = self.get_linode(
            kwargs={"linode_id": kwargs.get("linode_id", None), "name": name}
        )
        config_id = kwargs.get("config_id", None)
        check_running = kwargs.get("check_running", True)
        linode_id = instance.get("id", None)
        name = instance.get("label", None)

        if check_running:
            if instance["status"] == "running":
                raise SaltCloudSystemExit(
                    "Cannot boot Linode {0} ({1}). "
                    "Linode {0} is already running.".format(name, linode_id)
                )

        self._query(
            f"/linode/instances/{linode_id}/boot",
            method="POST",
            data={"config_id": config_id},
        )

        self._wait_for_linode_status(linode_id, "running")
        return True

    def clone(self, kwargs=None):
        linode_id = kwargs.get("linode_id", None)
        location = kwargs.get("location", None)
        size = kwargs.get("size", None)

        for item in [linode_id, location, size]:
            if item is None:
                raise SaltCloudSystemExit(
                    "The clone function requires a 'linode_id', 'location',"
                    "and 'size' to be provided."
                )

        return self._query(
            f"/linode/instances/{linode_id}/clone",
            method="POST",
            data={"region": location, "type": size},
        )

    def create_config(self, kwargs=None):
        name = kwargs.get("name", None)
        linode_id = kwargs.get("linode_id", None)
        root_disk_id = kwargs.get("root_disk_id", None)
        swap_disk_id = kwargs.get("swap_disk_id", None)
        data_disk_id = kwargs.get("data_disk_id", None)

        if not name and not linode_id:
            raise SaltCloudSystemExit(
                "The create_config function requires either a 'name' or 'linode_id'"
            )

        required_params = [name, linode_id, root_disk_id, swap_disk_id]
        for item in required_params:
            if item is None:
                raise SaltCloudSystemExit(
                    "The create_config functions requires a 'name', 'linode_id', "
                    "'root_disk_id', and 'swap_disk_id'."
                )

        devices = {
            "sda": {"disk_id": int(root_disk_id)},
            "sdb": {"disk_id": int(data_disk_id)} if data_disk_id is not None else None,
            "sdc": {"disk_id": int(swap_disk_id)},
        }

        return self._query(
            f"/linode/instances/{linode_id}/configs",
            method="POST",
            data={"label": name, "devices": devices},
        )

    def create(self, vm_):
        name = vm_["name"]

        if not _validate_name(name):
            return False

        __utils__["cloud.fire_event"](
            "event",
            "starting create",
            f"salt/cloud/{name}/creating",
            args=__utils__["cloud.filter_event"](
                "creating", vm_, ["name", "profile", "provider", "driver"]
            ),
            sock_dir=__opts__["sock_dir"],
            transport=__opts__["transport"],
        )

        log.info("Creating Cloud VM %s", name)

        result = None

        pub_ssh_keys = _get_ssh_keys(vm_)
        ssh_interface = _get_ssh_interface(vm_)
        use_private_ip = ssh_interface == "private_ips"
        assign_private_ip = _get_private_ip(vm_) or use_private_ip
        password = _get_password(vm_)
        swap_size = _get_swap_size(vm_)
        backups_enabled = _get_backup_enabled(vm_)

        clonefrom_name = vm_.get("clonefrom", None)
        instance_type = vm_.get("size", None)
        image = vm_.get("image", None)
        should_clone = True if clonefrom_name else False

        if should_clone:
            # clone into new linode
            clone_linode = self.get_linode(kwargs={"name": clonefrom_name})
            result = clone(
                {
                    "linode_id": clone_linode["id"],
                    "location": clone_linode["region"],
                    "size": clone_linode["type"],
                }
            )

            # create private IP if needed
            if assign_private_ip:
                self._query(
                    "/networking/ips",
                    method="POST",
                    data={"type": "ipv4", "public": False, "linode_id": result["id"]},
                )
        else:
            # create new linode
            result = self._query(
                "/linode/instances",
                method="POST",
                data={
                    "backups_enabled": backups_enabled,
                    "label": name,
                    "type": instance_type,
                    "region": vm_.get("location", None),
                    "private_ip": assign_private_ip,
                    "booted": True,
                    "root_pass": password,
                    "authorized_keys": pub_ssh_keys,
                    "image": image,
                    "swap_size": swap_size,
                },
            )

        linode_id = result.get("id", None)

        # wait for linode to be created
        self._wait_for_event("linode_create", "linode", linode_id, "finished")
        log.debug("linode '%s' has been created", name)

        if should_clone:
            self.boot(kwargs={"linode_id": linode_id})

        # wait for linode to finish booting
        self._wait_for_linode_status(linode_id, "running")

        public_ips, private_ips = self._get_ips(linode_id)

        data = {}
        data["id"] = linode_id
        data["name"] = result["label"]
        data["size"] = result["type"]
        data["state"] = result["status"]
        data["ipv4"] = result["ipv4"]
        data["ipv6"] = result["ipv6"]
        data["public_ips"] = public_ips
        data["private_ips"] = private_ips

        if use_private_ip:
            vm_["ssh_host"] = private_ips[0]
        else:
            vm_["ssh_host"] = public_ips[0]

        # Send event that the instance has booted.
        __utils__["cloud.fire_event"](
            "event",
            "waiting for ssh",
            f"salt/cloud/{name}/waiting_for_ssh",
            sock_dir=__opts__["sock_dir"],
            args={"ip_address": vm_["ssh_host"]},
            transport=__opts__["transport"],
        )

        ret = __utils__["cloud.bootstrap"](vm_, __opts__)
        ret.update(data)

        log.info("Created Cloud VM '%s'", name)
        log.debug("'%s' VM creation details:\n%s", name, pprint.pformat(data))

        __utils__["cloud.fire_event"](
            "event",
            "created instance",
            f"salt/cloud/{name}/created",
            args=__utils__["cloud.filter_event"](
                "created", vm_, ["name", "profile", "provider", "driver"]
            ),
            sock_dir=__opts__["sock_dir"],
            transport=__opts__["transport"],
        )

        return ret

    def destroy(self, name):
        __utils__["cloud.fire_event"](
            "event",
            "destroyed instance",
            f"salt/cloud/{name}/destroyed",
            args={"name": name},
            sock_dir=__opts__["sock_dir"],
            transport=__opts__["transport"],
        )

        if __opts__.get("update_cachedir", False) is True:
            __utils__["cloud.delete_minion_cachedir"](
                name, _get_active_provider_name().split(":")[0], __opts__
            )

        instance = self._get_linode_by_name(name)
        linode_id = instance.get("id", None)

        self._query(f"/linode/instances/{linode_id}", method="DELETE")

    def get_config_id(self, kwargs=None):
        name = kwargs.get("name", None)
        linode_id = kwargs.get("linode_id", None)

        if name is None and linode_id is None:
            raise SaltCloudSystemExit(
                "The get_config_id function requires either a 'name' or a 'linode_id' "
                "to be provided."
            )

        if linode_id is None:
            linode_id = self.get_linode(kwargs=kwargs).get("id", None)

        response = self._query(f"/linode/instances/{linode_id}/configs")
        configs = response.get("data", [])

        return {"config_id": configs[0]["id"]}

    def list_nodes_min(self):
        result = self._query("/linode/instances")
        instances = result.get("data", [])

        ret = {}
        for instance in instances:
            name = instance["label"]
            ret[name] = {"id": instance["id"], "state": instance["status"]}

        return ret

    def list_nodes_full(self):
        return self._list_linodes(full=True)

    def list_nodes(self):
        return self._list_linodes()

    def reboot(self, name):
        instance = self._get_linode_by_name(name)
        linode_id = instance.get("id", None)

        self._query(f"/linode/instances/{linode_id}/reboot", method="POST")
        return self._wait_for_linode_status(linode_id, "running")

    def show_instance(self, name):
        instance = self._get_linode_by_name(name)
        linode_id = instance.get("id", None)
        public_ips, private_ips = self._get_ips(linode_id)

        return {
            "id": instance["id"],
            "image": instance["image"],
            "name": instance["label"],
            "size": instance["type"],
            "state": instance["status"],
            "public_ips": public_ips,
            "private_ips": private_ips,
        }

    def show_pricing(self, kwargs=None):
        profile = __opts__["profiles"].get(kwargs["profile"], {})
        if not profile:
            raise SaltCloudNotFound("The requested profile was not found.")

        # Make sure the profile belongs to Linode
        provider = profile.get("provider", "0:0")
        comps = provider.split(":")
        if len(comps) < 2 or comps[1] != "linode":
            raise SaltCloudException("The requested profile does not belong to Linode.")

        instance_type = self._get_linode_type(profile["size"])
        pricing = instance_type.get("price", {})

        per_hour = pricing["hourly"]
        per_day = per_hour * 24
        per_week = per_day * 7
        per_month = pricing["monthly"]
        per_year = per_month * 12

        return {
            profile["profile"]: {
                "per_hour": per_hour,
                "per_day": per_day,
                "per_week": per_week,
                "per_month": per_month,
                "per_year": per_year,
            }
        }

    def start(self, name):
        instance = self._get_linode_by_name(name)
        linode_id = instance.get("id", None)

        if instance["status"] == "running":
            return {
                "success": True,
                "action": "start",
                "state": "Running",
                "msg": "Machine already running",
            }

        self._query(f"/linode/instances/{linode_id}/boot", method="POST")

        self._wait_for_linode_status(linode_id, "running")
        return {
            "success": True,
            "state": "Running",
            "action": "start",
        }

    def stop(self, name):
        instance = self._get_linode_by_name(name)
        linode_id = instance.get("id", None)

        if instance["status"] == "offline":
            return {
                "success": True,
                "action": "stop",
                "state": "Stopped",
                "msg": "Machine already stopped",
            }

        self._query(f"/linode/instances/{linode_id}/shutdown", method="POST")

        self._wait_for_linode_status(linode_id, "offline")
        return {"success": True, "state": "Stopped", "action": "stop"}

    def _get_linode_by_id(self, linode_id):
        return self._query(f"/linode/instances/{linode_id}")

    def _get_linode_by_name(self, name):
        result = self._query("/linode/instances")
        instances = result.get("data", [])

        for instance in instances:
            if instance["label"] == name:
                return instance

        raise SaltCloudNotFound(f"The specified name, {name}, could not be found.")

    def _list_linodes(self, full=False):
        result = self._query("/linode/instances")
        instances = result.get("data", [])

        ret = {}
        for instance in instances:
            node = {}
            node["id"] = instance["id"]
            node["image"] = instance["image"]
            node["name"] = instance["label"]
            node["size"] = instance["type"]
            node["state"] = instance["status"]

            public_ips, private_ips = self._get_ips(node["id"])
            node["public_ips"] = public_ips
            node["private_ips"] = private_ips

            if full:
                node["extra"] = instance

            ret[instance["label"]] = node

        return ret

    def _get_linode_type(self, linode_type):
        return self._query(f"/linode/types/{linode_type}")

    def _get_ips(self, linode_id):
        instance = self._get_linode_by_id(linode_id)
        public = []
        private = []

        for addr in instance.get("ipv4", []):
            if ipaddress.ip_address(addr).is_private:
                private.append(addr)
            else:
                public.append(addr)

        return (public, private)

    def _poll(
        self,
        description,
        getter,
        condition,
        timeout=None,
        poll_interval=None,
    ):
        """
        Return true in handler to signal complete.
        """
        if poll_interval is None:
            poll_interval = _get_poll_interval()

        if timeout is None:
            timeout = 120

        times = (timeout * 1000) / poll_interval
        curr = 0

        while True:
            curr += 1
            result = getter()
            if condition(result):
                return True
            elif curr <= times:
                time.sleep(poll_interval / 1000)
                log.info("retrying: polling for %s...", description)
            else:
                raise SaltCloudException(f"timed out: polling for {description}")

    def _wait_for_entity_status(
        self, getter, status, entity_name="item", identifier="some", timeout=None
    ):
        return self._poll(
            f"{entity_name} (id={identifier}) status to be '{status}'",
            getter,
            lambda item: item.get("status") == status,
            timeout=timeout,
        )

    def _wait_for_linode_status(self, linode_id, status, timeout=None):
        return self._wait_for_entity_status(
            lambda: self._get_linode_by_id(linode_id),
            status,
            entity_name="linode",
            identifier=linode_id,
            timeout=timeout,
        )

    def _check_event_status(self, event, desired_status):
        status = event.get("status")
        action = event.get("action")
        entity = event.get("entity")
        if status == "failed":
            raise SaltCloudSystemExit(
                "event {} for {} (id={}) failed".format(
                    action, entity["type"], entity["id"]
                )
            )
        return status == desired_status

    def _wait_for_event(self, action, entity, entity_id, status, timeout=None):
        event_filter = {
            "+order_by": "created",
            "+order": "desc",
            "seen": False,
            "action": action,
            "entity.id": entity_id,
            "entity.type": entity,
        }
        last_event = None

        def condition(event):
            return self._check_event_status(event, status)

        while True:
            if last_event is not None:
                event_filter["+gt"] = last_event
            filter_json = json.dumps(event_filter, separators=(",", ":"))
            result = self._query("/account/events", headers={"X-Filter": filter_json})
            events = result.get("data", [])

            if len(events) == 0:
                break

            for event in events:
                event_id = event.get("id")
                event_entity = event.get("entity", None)
                last_event = event_id
                if not event_entity:
                    continue

                if not (
                    event_entity["type"] == entity
                    and event_entity["id"] == entity_id
                    and event.get("action") == action
                ):
                    continue

                if condition(event):
                    return True

                return self._poll(
                    f"event {event_id} to be '{status}'",
                    lambda: self._query(f"/account/events/{event_id}"),
                    condition,
                    timeout=timeout,
                )

        return False

    def _get_response_json(self, response):
        json = None
        try:
            json = response.json()
        except ValueError:
            pass
        return json


def avail_images(call=None):
    """
    Return available Linode images.

    CLI Example:

    .. code-block:: bash

        salt-cloud --list-images my-linode-config
        salt-cloud -f avail_images my-linode-config
    """
    if call == "action":
        raise SaltCloudException(
            "The avail_images function must be called with -f or --function."
        )
    return LinodeAPIv4.get_api_instance().avail_images()


def avail_locations(call=None):
    """
    Return available Linode datacenter locations.

    CLI Example:

    .. code-block:: bash

        salt-cloud --list-locations my-linode-config
        salt-cloud -f avail_locations my-linode-config
    """
    if call == "action":
        raise SaltCloudException(
            "The avail_locations function must be called with -f or --function."
        )
    return LinodeAPIv4.get_api_instance().avail_locations()


def avail_sizes(call=None):
    """
    Return available Linode sizes.

    CLI Example:

    .. code-block:: bash

        salt-cloud --list-sizes my-linode-config
        salt-cloud -f avail_sizes my-linode-config
    """
    if call == "action":
        raise SaltCloudException(
            "The avail_locations function must be called with -f or --function."
        )
    return LinodeAPIv4.get_api_instance().avail_sizes()


def set_backup_schedule(name=None, kwargs=None, call=None):
    """
    Set the backup schedule for a Linode.

    name
        The name (label) of the Linode. Can be used instead of
        ``linode_id``.

    linode_id
        The ID of the Linode instance to set the backup schedule for.
        If provided, will be used as an alternative to ``name`` and
        reduces the number of API calls to Linode by one. Will be
        preferred over ``name``.

    auto_enable
        If ``True``, automatically enable the backup feature for the Linode
        if it wasn't already enabled. Optional parameter, default to ``False``.

    day
        Possible values:
        ``Sunday``, ``Monday``, ``Tuesday``, ``Wednesday``,
        ``Thursday``, ``Friday``, ``Saturday``

        The day of the week that your Linode's weekly Backup is taken.
        If not set manually, a day will be chosen for you. Backups are
        taken every day, but backups taken on this day are preferred
        when selecting backups to retain for a longer period.

        If not set manually, then when backups are initially enabled,
        this may come back as ``Scheduling`` until the day is automatically
        selected.

    window
        Possible values:
        ``W0``, ``W2``, ``W4``, ``W6``, ``W8``, ``W10``,
        ``W12``, ``W14``, ``W16``, ``W18``, ``W20``, ``W22``

        The window in which your backups will be taken, in UTC. A backups
        window is a two-hour span of time in which the backup may occur.

        For example, ``W10`` indicates that your backups should be taken
        between 10:00 and 12:00. If you do not choose a backup window, one
        will be selected for you automatically.

        If not set manually, when backups are initially enabled this may come
        back as ``Scheduling`` until the window is automatically selected.

    Can be called as an action (which requires a name):

    .. code-block:: bash

        salt-cloud -a set_backup_schedule my-linode-instance day=Monday window=W20 auto_enable=True

    ...or as a function (which requires either a name or linode_id):

    .. code-block:: bash

        salt-cloud -f set_backup_schedule my-linode-provider name=my-linode-instance day=Monday window=W20 auto_enable=True
        salt-cloud -f set_backup_schedule my-linode-provider linode_id=1225876 day=Monday window=W20 auto_enable=True
    """
    if name is None and call == "action":
        raise SaltCloudSystemExit(
            "The set_backup_schedule backup schedule "
            "action requires the name of the Linode.",
        )

    if kwargs is None:
        kwargs = {}

    if call == "function":
        name = kwargs.get("name", None)
    linode_id = kwargs.get("linode_id")

    auto_enable = str(kwargs.get("auto_enable")).lower() == "true"

    if name is None and linode_id is None:
        raise SaltCloudSystemExit(
            "The set_backup_schedule function requires "
            "either a 'name' or a 'linode_id'."
        )

    return LinodeAPIv4.get_api_instance().set_backup_schedule(
        day=kwargs.get("day"),
        window=kwargs.get("window"),
        label=name,
        linode_id=linode_id,
        auto_enable=auto_enable,
    )


def boot(name=None, kwargs=None, call=None):
    """
    Boot a Linode.

    name
        The name of the Linode to boot. Can be used instead of ``linode_id``.

    linode_id
        The ID of the Linode to boot. If provided, will be used as an
        alternative to ``name`` and reduces the number of API calls to
        Linode by one. Will be preferred over ``name``.

    config_id
        The ID of the Config to boot. Required.

    check_running
        Defaults to True. If set to False, overrides the call to check if
        the VM is running before calling the linode.boot API call. Change
        ``check_running`` to True is useful during the boot call in the
        create function, since the new VM will not be running yet.

    Can be called as an action (which requires a name):

    .. code-block:: bash

        salt-cloud -a boot my-instance config_id=10

    ...or as a function (which requires either a name or linode_id):

    .. code-block:: bash

        salt-cloud -f boot my-linode-config name=my-instance config_id=10
        salt-cloud -f boot my-linode-config linode_id=1225876 config_id=10
    """
    if name is None and call == "action":
        raise SaltCloudSystemExit("The boot action requires a 'name'.")

    linode_id = kwargs.get("linode_id", None)
    config_id = kwargs.get("config_id", None)

    if call == "function":
        name = kwargs.get("name", None)

    if name is None and linode_id is None:
        raise SaltCloudSystemExit(
            "The boot function requires either a 'name' or a 'linode_id'."
        )

    return LinodeAPIv4.get_api_instance().boot(name=name, kwargs=kwargs)


def clone(kwargs=None, call=None):
    """
    Clone a Linode.

    linode_id
        The ID of the Linode to clone. Required.

    location
        The location of the new Linode. Required.

    size
        The size of the new Linode (must be greater than or equal to the clone source). Required.

    CLI Example:

    .. code-block:: bash

        salt-cloud -f clone my-linode-config linode_id=1234567 location=us-central size=g6-standard-1
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The clone function must be called with -f or --function."
        )

    return LinodeAPIv4.get_api_instance().clone(kwargs=kwargs)


def create(vm_):
    """
    Create a single Linode VM.
    """
    try:
        # Check for required profile parameters before sending any API calls.
        if (
            vm_["profile"]
            and config.is_profile_configured(
                __opts__,
                _get_active_provider_name() or "linode",
                vm_["profile"],
                vm_=vm_,
            )
        ) is False:
            return False
    except AttributeError:
        pass

    return LinodeAPIv4.get_api_instance().create(vm_)


def create_config(kwargs=None, call=None):
    """
    Creates a Linode Configuration Profile.

    name
        The name of the VM to create the config for.

    linode_id
        The ID of the Linode to create the configuration for.

    root_disk_id
        The Root Disk ID to be used for this config.

    swap_disk_id
        The Swap Disk ID to be used for this config.

    data_disk_id
        The Data Disk ID to be used for this config.

    .. versionadded:: 2016.3.0

    kernel_id
        The ID of the kernel to use for this configuration profile.
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The create_config function must be called with -f or --function."
        )
    return LinodeAPIv4.get_api_instance().create_config(kwargs=kwargs)


def destroy(name, call=None):
    """
    Destroys a Linode by name.

    name
        The name of VM to be be destroyed.

    CLI Example:

    .. code-block:: bash

        salt-cloud -d vm_name
    """
    if call == "function":
        raise SaltCloudException(
            "The destroy action must be called with -d, --destroy, -a or --action."
        )
    return LinodeAPIv4.get_api_instance().destroy(name)


def get_config_id(kwargs=None, call=None):
    """
    Returns a config_id for a given linode.

    .. versionadded:: 2015.8.0

    name
        The name of the Linode for which to get the config_id. Can be used instead
        of ``linode_id``.

    linode_id
        The ID of the Linode for which to get the config_id. Can be used instead
        of ``name``.

    CLI Example:

    .. code-block:: bash

        salt-cloud -f get_config_id my-linode-config name=my-linode
        salt-cloud -f get_config_id my-linode-config linode_id=1234567
    """
    if call == "action":
        raise SaltCloudException(
            "The get_config_id function must be called with -f or --function."
        )
    return LinodeAPIv4.get_api_instance().get_config_id(kwargs=kwargs)


def get_linode(kwargs=None, call=None):
    """
    Returns data for a single named Linode.

    name
        The name of the Linode for which to get data. Can be used instead
        ``linode_id``. Note this will induce an additional API call
        compared to using ``linode_id``.

    linode_id
        The ID of the Linode for which to get data. Can be used instead of
        ``name``.

    CLI Example:

    .. code-block:: bash

        salt-cloud -f get_linode my-linode-config name=my-instance
        salt-cloud -f get_linode my-linode-config linode_id=1234567
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The get_linode function must be called with -f or --function."
        )
    return LinodeAPIv4.get_api_instance().get_linode(kwargs=kwargs)


def list_nodes(call=None):
    """
    Returns a list of linodes, keeping only a brief listing.

    CLI Example:

    .. code-block:: bash

        salt-cloud -Q
        salt-cloud --query
        salt-cloud -f list_nodes my-linode-config

    .. note::

        The ``image`` label only displays information about the VM's distribution vendor,
        such as "Debian" or "RHEL" and does not display the actual image name. This is
        due to a limitation of the Linode API.
    """
    if call == "action":
        raise SaltCloudException(
            "The list_nodes function must be called with -f or --function."
        )
    return LinodeAPIv4.get_api_instance().list_nodes()


def list_nodes_full(call=None):
    """
    List linodes, with all available information.

    CLI Example:

    .. code-block:: bash

        salt-cloud -F
        salt-cloud --full-query
        salt-cloud -f list_nodes_full my-linode-config

    .. note::

        The ``image`` label only displays information about the VM's distribution vendor,
        such as "Debian" or "RHEL" and does not display the actual image name. This is
        due to a limitation of the Linode API.
    """
    if call == "action":
        raise SaltCloudException(
            "The list_nodes_full function must be called with -f or --function."
        )
    return LinodeAPIv4.get_api_instance().list_nodes_full()


def list_nodes_min(call=None):
    """
    Return a list of the VMs that are on the provider. Only a list of VM names and
    their state is returned. This is the minimum amount of information needed to
    check for existing VMs.

    .. versionadded:: 2015.8.0

    CLI Example:

    .. code-block:: bash

        salt-cloud -f list_nodes_min my-linode-config
        salt-cloud --function list_nodes_min my-linode-config
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The list_nodes_min function must be called with -f or --function."
        )
    return LinodeAPIv4.get_api_instance().list_nodes_min()


def list_nodes_select(call=None):
    """
    Return a list of the VMs that are on the provider, with select fields.
    """
    return LinodeAPIv4.get_api_instance().list_nodes_select(call)


def reboot(name, call=None):
    """
    Reboot a linode.

    .. versionadded:: 2015.8.0

    name
        The name of the VM to reboot.

    CLI Example:

    .. code-block:: bash

        salt-cloud -a reboot vm_name
    """
    if call != "action":
        raise SaltCloudException(
            "The show_instance action must be called with -a or --action."
        )
    return LinodeAPIv4.get_api_instance().reboot(name)


def show_instance(name, call=None):
    """
    Displays details about a particular Linode VM. Either a name or a linode_id must
    be provided.

    .. versionadded:: 2015.8.0

    name
        The name of the VM for which to display details.

    CLI Example:

    .. code-block:: bash

        salt-cloud -a show_instance vm_name

    .. note::

        The ``image`` label only displays information about the VM's distribution vendor,
        such as "Debian" or "RHEL" and does not display the actual image name. This is
        due to a limitation of the Linode API.
    """
    if call != "action":
        raise SaltCloudException(
            "The show_instance action must be called with -a or --action."
        )
    return LinodeAPIv4.get_api_instance().show_instance(name)


def show_pricing(kwargs=None, call=None):
    """
    Show pricing for a particular profile. This is only an estimate, based on
    unofficial pricing sources.

    .. versionadded:: 2015.8.0

    CLI Example:

    .. code-block:: bash

        salt-cloud -f show_pricing my-linode-config profile=my-linode-profile
    """
    if call != "function":
        raise SaltCloudException(
            "The show_instance action must be called with -f or --function."
        )
    return LinodeAPIv4.get_api_instance().show_pricing(kwargs=kwargs)


def start(name, call=None):
    """
    Start a VM in Linode.

    name
        The name of the VM to start.

    CLI Example:

    .. code-block:: bash

        salt-cloud -a stop vm_name
    """
    if call != "action":
        raise SaltCloudException("The start action must be called with -a or --action.")
    return LinodeAPIv4.get_api_instance().start(name)


def stop(name, call=None):
    """
    Stop a VM in Linode.

    name
        The name of the VM to stop.

    CLI Example:

    .. code-block:: bash

        salt-cloud -a stop vm_name
    """
    if call != "action":
        raise SaltCloudException("The stop action must be called with -a or --action.")
    return LinodeAPIv4.get_api_instance().stop(name)



"""
    :codeauthor: Hristo Voyvodov <hristo.voyvodov@redsift.io>
"""

import pytest

import salt.states.firewalld as firewalld
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {firewalld: {"__opts__": {"test": False}}}


def test_masquerade():
    firewalld_reload_rules = MagicMock(return_value={})
    firewalld_get_zones = MagicMock(
        return_value=[
            "block",
            "public",
        ]
    )
    firewalld_get_masquerade = MagicMock(return_value=True)
    firewalld_remove_masquerade = MagicMock(return_value={})

    __salt__ = {
        "firewalld.reload_rules": firewalld_reload_rules,
        "firewalld.get_zones": firewalld_get_zones,
        "firewalld.get_masquerade": firewalld_get_masquerade,
        "firewalld.remove_masquerade": firewalld_remove_masquerade,
    }
    with patch.dict(firewalld.__dict__, {"__salt__": __salt__}):
        ret = firewalld.present("public")
        assert ret == {
            "changes": {},
            "result": True,
            "comment": "'public' is already in the desired state.",
            "name": "public",
        }

        ret = firewalld.present("public", masquerade=False)
        assert ret == {
            "changes": {
                "masquerade": {
                    "old": True,
                    "new": "Masquerading successfully disabled.",
                }
            },
            "result": True,
            "comment": "'public' was configured.",
            "name": "public",
        }

        ret = firewalld.present("public", masquerade=True)
        assert ret == {
            "changes": {},
            "result": True,
            "comment": "'public' is already in the desired state.",
            "name": "public",
        }



"""
Test cases for salt.client.ssh.wrapper.defaults

This has been copied 1:1 from tests.pytests.unit.modules.test_defaults
"""

import inspect

import pytest

import salt.client.ssh.wrapper.defaults as defaults
from tests.support.mock import MagicMock, patch


@pytest.fixture()
def configure_loader_modules():
    return {defaults: {}}


def test_get_mock():
    """
    Test if it execute a defaults client run and return a dict
    """
    with patch.object(inspect, "stack", MagicMock(return_value=[])), patch(
        "salt.client.ssh.wrapper.defaults.get",
        MagicMock(return_value={"users": {"root": [0]}}),
    ):
        assert defaults.get("core:users:root") == {"users": {"root": [0]}}


def test_merge_with_list_merging():
    """
    Test deep merging of dicts with merge_lists enabled.
    """

    src_dict = {
        "string_key": "string_val_src",
        "list_key": ["list_val_src"],
        "dict_key": {"dict_key_src": "dict_val_src"},
    }

    dest_dict = {
        "string_key": "string_val_dest",
        "list_key": ["list_val_dest"],
        "dict_key": {"dict_key_dest": "dict_val_dest"},
    }

    merged_dict = {
        "string_key": "string_val_src",
        "list_key": ["list_val_dest", "list_val_src"],
        "dict_key": {
            "dict_key_dest": "dict_val_dest",
            "dict_key_src": "dict_val_src",
        },
    }

    defaults.merge(dest_dict, src_dict, merge_lists=True)
    assert dest_dict == merged_dict


def test_merge_without_list_merging():
    """
    Test deep merging of dicts with merge_lists disabled.
    """

    src = {
        "string_key": "string_val_src",
        "list_key": ["list_val_src"],
        "dict_key": {"dict_key_src": "dict_val_src"},
    }

    dest = {
        "string_key": "string_val_dest",
        "list_key": ["list_val_dest"],
        "dict_key": {"dict_key_dest": "dict_val_dest"},
    }

    merged = {
        "string_key": "string_val_src",
        "list_key": ["list_val_src"],
        "dict_key": {
            "dict_key_dest": "dict_val_dest",
            "dict_key_src": "dict_val_src",
        },
    }

    defaults.merge(dest, src, merge_lists=False)
    assert dest == merged


def test_merge_not_in_place():
    """
    Test deep merging of dicts not in place.
    """

    src = {"nested_dict": {"A": "A"}}

    dest = {"nested_dict": {"B": "B"}}

    dest_orig = {"nested_dict": {"B": "B"}}

    merged = {"nested_dict": {"A": "A", "B": "B"}}

    final = defaults.merge(dest, src, in_place=False)
    assert dest == dest_orig
    assert final == merged


def test_merge_src_is_none():
    """
    Test deep merging of dicts not in place.
    """

    dest = {"nested_dict": {"B": "B"}}

    dest_orig = {"nested_dict": {"B": "B"}}

    final = defaults.merge(dest, None, in_place=False)
    assert dest == dest_orig
    assert final == dest_orig


def test_merge_dest_is_none():
    """
    Test deep merging of dicts not in place.
    """

    src = {"nested_dict": {"B": "B"}}

    src_orig = {"nested_dict": {"B": "B"}}

    final = defaults.merge(None, src, in_place=False)
    assert src == src_orig
    assert final == src_orig


def test_merge_in_place_dest_is_none():
    """
    Test deep merging of dicts not in place.
    """

    src = {"nested_dict": {"B": "B"}}

    pytest.raises(TypeError, defaults.merge, None, src)


def test_deepcopy():
    """
    Test a deep copy of object.
    """

    src = {"A": "A", "B": "B"}

    dist = defaults.deepcopy(src)
    dist.update({"C": "C"})

    result = {"A": "A", "B": "B", "C": "C"}

    assert src != dist
    assert dist == result


def test_update_in_place():
    """
    Test update with defaults values in place.
    """

    group01 = {
        "defaults": {"enabled": True, "extra": ["test", "stage"]},
        "nodes": {"host01": {"index": "foo", "upstream": "bar"}},
    }

    host01 = {
        "enabled": True,
        "index": "foo",
        "upstream": "bar",
        "extra": ["test", "stage"],
    }

    defaults.update(group01["nodes"], group01["defaults"])
    assert group01["nodes"]["host01"] == host01


def test_update_with_defaults_none():
    group01 = {
        "defaults": {"enabled": True, "extra": ["test", "stage"]},
        "nodes": {"host01": {"index": "foo", "upstream": "bar"}},
    }

    host01 = {
        "index": "foo",
        "upstream": "bar",
    }

    defaults.update(group01["nodes"], None)
    assert group01["nodes"]["host01"] == host01


def test_update_with_dest_none():
    group01 = {
        "defaults": {"enabled": True, "extra": ["test", "stage"]},
        "nodes": {"host01": {"index": "foo", "upstream": "bar"}},
    }

    ret = defaults.update(None, group01["defaults"], in_place=False)
    assert ret == {}


def test_update_in_place_with_dest_none():
    group01 = {
        "defaults": {"enabled": True, "extra": ["test", "stage"]},
        "nodes": {"host01": {"index": "foo", "upstream": "bar"}},
    }

    pytest.raises(TypeError, defaults.update, None, group01["defaults"])



"""
integration tests for mac_xattr
"""

import pytest

from salt.exceptions import CommandExecutionError

pytestmark = [
    pytest.mark.skip_if_binaries_missing("xattr"),
    pytest.mark.slow_test,
    pytest.mark.skip_if_not_root,
    pytest.mark.skip_unless_on_darwin,
]


@pytest.fixture(scope="module")
def xattr(modules):
    return modules.xattr


@pytest.fixture
def existing_file(tmp_path):
    fpath = tmp_path / "xattr_test_file.txt"
    fpath.touch()
    return fpath


@pytest.fixture
def non_existing_file(tmp_path):
    return tmp_path / "xattr_no_file"


def test_list_no_xattr(xattr, existing_file, non_existing_file):
    """
    Make sure there are no attributes
    """
    # Clear existing attributes
    ret = xattr.clear(existing_file)
    assert ret

    # Test no attributes
    ret = xattr.list(existing_file)
    assert ret == {}

    # Test file not found
    with pytest.raises(CommandExecutionError) as exc:
        ret = xattr.list(non_existing_file)
        assert f"File not found: {non_existing_file}" in str(exc.value)


def test_write(xattr, existing_file, non_existing_file):
    """
    Write an attribute
    """
    # Clear existing attributes
    ret = xattr.clear(existing_file)
    assert ret

    # Write some attributes
    ret = xattr.write(existing_file, "spongebob", "squarepants")
    assert ret

    ret = xattr.write(existing_file, "squidward", "plankton")
    assert ret

    ret = xattr.write(existing_file, "crabby", "patty")
    assert ret

    # Test that they were actually added
    ret = xattr.list(existing_file)
    assert ret == {
        "spongebob": "squarepants",
        "squidward": "plankton",
        "crabby": "patty",
    }

    # Test file not found
    with pytest.raises(CommandExecutionError) as exc:
        ret = xattr.write(non_existing_file, "patrick", "jellyfish")
        assert f"File not found: {non_existing_file}" in str(exc.value)


def test_read(xattr, existing_file, non_existing_file):
    """
    Test xattr.read
    """
    # Clear existing attributes
    ret = xattr.clear(existing_file)
    assert ret

    # Write an attribute
    ret = xattr.write(existing_file, "spongebob", "squarepants")
    assert ret

    # Read the attribute
    ret = xattr.read(existing_file, "spongebob")
    assert ret == "squarepants"

    # Test file not found
    with pytest.raises(CommandExecutionError) as exc:
        ret = xattr.read(non_existing_file, "spongebob")
        assert f"File not found: {non_existing_file}" in str(exc.value)

    # Test attribute not found
    with pytest.raises(CommandExecutionError) as exc:
        ret = xattr.read(existing_file, "patrick")
        assert "Attribute not found: patrick" in str(exc.value)


def test_delete(xattr, existing_file, non_existing_file):
    """
    Test xattr.delete
    """
    # Clear existing attributes
    ret = xattr.clear(existing_file)
    assert ret

    # Write some attributes
    ret = xattr.write(existing_file, "spongebob", "squarepants")
    assert ret

    ret = xattr.write(existing_file, "squidward", "plankton")
    assert ret

    ret = xattr.write(existing_file, "crabby", "patty")
    assert ret

    # Delete an attribute
    ret = xattr.delete(existing_file, "squidward")
    assert ret

    # Make sure it was actually deleted
    ret = xattr.list(existing_file)
    assert ret == {
        "spongebob": "squarepants",
        "crabby": "patty",
    }

    # Test file not found
    with pytest.raises(CommandExecutionError) as exc:
        ret = xattr.delete(non_existing_file, "spongebob")
        assert f"File not found: {non_existing_file}" in str(exc.value)

    # Test attribute not found
    with pytest.raises(CommandExecutionError) as exc:
        ret = xattr.delete(existing_file, "patrick")
        assert "Attribute not found: patrick" in str(exc.value)


def test_clear(xattr, existing_file, non_existing_file):
    """
    Test xattr.clear
    """
    # Clear existing attributes
    ret = xattr.clear(existing_file)
    assert ret

    # Write some attributes
    ret = xattr.write(existing_file, "spongebob", "squarepants")
    assert ret

    ret = xattr.write(existing_file, "squidward", "plankton")
    assert ret

    ret = xattr.write(existing_file, "crabby", "patty")
    assert ret

    # Test Clear
    ret = xattr.clear(existing_file)
    assert ret

    # Test file not found
    with pytest.raises(CommandExecutionError) as exc:
        ret = xattr.clear(non_existing_file)
        assert f"File not found: {non_existing_file}" in str(exc.value)



import pathlib

import pytest

import salt.crypt


@pytest.mark.windows_whitelisted
def test_generated_keys(tmp_path):
    priv = pathlib.Path(salt.crypt.gen_keys(tmp_path, "aaa", 2048))
    pub = priv.with_suffix(".pub")
    assert "\r" not in priv.read_text(encoding="utf-8")
    assert "\r" not in pub.read_text(encoding="utf-8")



"""
Integration tests for Ruby Gem module
"""

import pytest
from tornado.httpclient import HTTPClient

import salt.utils.platform
from tests.support.case import ModuleCase


def check_status():
    """
    Check the status of the rubygems source
    """
    try:
        return HTTPClient().fetch("https://rubygems.org").code == 200
    except Exception:  # pylint: disable=broad-except
        return False


@pytest.mark.timeout_unless_on_windows(120)
@pytest.mark.skip_if_binaries_missing("gem")
@pytest.mark.windows_whitelisted
@pytest.mark.destructive_test
class GemModuleTest(ModuleCase):
    """
    Validate gem module
    """

    def setUp(self):
        if check_status() is False:
            self.skipTest("External resource 'https://rubygems.org' is not available")

        self.GEM_BIN = "gem.cmd" if salt.utils.platform.is_windows() else "gem"
        self.GEM = "tidy"
        self.GEM_VER = "1.1.2"
        self.OLD_GEM = "brass"
        self.OLD_VERSION = "1.0.0"
        self.NEW_VERSION = "1.2.1"
        self.GEM_LIST = [self.GEM, self.OLD_GEM]
        for name in (
            "GEM",
            "GEM_VER",
            "OLD_GEM",
            "OLD_VERSION",
            "NEW_VERSION",
            "GEM_LIST",
        ):
            self.addCleanup(delattr, self, name)

        def uninstall_gem():
            # Remove gem if it is already installed
            if self.run_function("gem.list", [self.GEM]):
                self.run_function("gem.uninstall", [self.GEM])

        self.addCleanup(uninstall_gem)

    def run_function(self, function, *args, **kwargs):
        """Override run_function to use the gem binary"""
        kwargs["gem_bin"] = self.GEM_BIN
        return super().run_function(function, *args, **kwargs)

    @pytest.mark.slow_test
    def test_install_uninstall(self):
        """
        gem.install
        gem.uninstall
        """
        self.run_function("gem.install", [self.GEM])
        gem_list = self.run_function("gem.list", [self.GEM])
        self.assertIn(self.GEM, gem_list)

        self.run_function("gem.uninstall", [self.GEM])
        self.assertFalse(self.run_function("gem.list", [self.GEM]))

    @pytest.mark.slow_test
    def test_install_version(self):
        """
        gem.install rake version=11.1.2
        """
        self.run_function("gem.install", [self.GEM], version=self.GEM_VER)
        gem_list = self.run_function("gem.list", [self.GEM])
        self.assertIn(self.GEM, gem_list)
        self.assertIn(self.GEM_VER, gem_list[self.GEM])

        self.run_function("gem.uninstall", [self.GEM])
        self.assertFalse(self.run_function("gem.list", [self.GEM]))

    @pytest.mark.slow_test
    def test_list(self):
        """
        gem.list
        """
        self.run_function("gem.install", [" ".join(self.GEM_LIST)])

        all_ret = self.run_function("gem.list")
        for gem in self.GEM_LIST:
            self.assertIn(gem, all_ret)

        single_ret = self.run_function("gem.list", [self.GEM])
        self.assertIn(self.GEM, single_ret)

        self.run_function("gem.uninstall", [" ".join(self.GEM_LIST)])

    @pytest.mark.slow_test
    def test_list_upgrades(self):
        """
        gem.list_upgrades
        """
        # install outdated gem
        self.run_function("gem.install", [self.OLD_GEM], version=self.OLD_VERSION)

        ret = self.run_function("gem.list_upgrades")
        self.assertIn(self.OLD_GEM, ret)

        self.run_function("gem.uninstall", [self.OLD_GEM])

    @pytest.mark.slow_test
    def test_sources_add_remove(self):
        """
        gem.sources_add
        gem.sources_remove
        """
        source = "http://production.cf.rubygems.org"

        self.run_function("gem.sources_add", [source])
        sources_list = self.run_function("gem.sources_list")
        self.assertIn(source, sources_list)

        self.run_function("gem.sources_remove", [source])
        sources_list = self.run_function("gem.sources_list")
        self.assertNotIn(source, sources_list)

    @pytest.mark.slow_test
    def test_update(self):
        """
        gem.update
        """
        self.run_function("gem.install", [self.OLD_GEM], version=self.OLD_VERSION)
        gem_list = self.run_function("gem.list", [self.OLD_GEM])
        self.assertEqual({self.OLD_GEM: [self.OLD_VERSION]}, gem_list)

        self.run_function("gem.update", [self.OLD_GEM])
        gem_list = self.run_function("gem.list", [self.OLD_GEM])
        self.assertEqual({self.OLD_GEM: [self.NEW_VERSION, self.OLD_VERSION]}, gem_list)

        self.run_function("gem.uninstall", [self.OLD_GEM])
        self.assertFalse(self.run_function("gem.list", [self.OLD_GEM]))

    @pytest.mark.slow_test
    def test_update_system(self):
        """
        gem.update_system
        """
        ret = self.run_function("gem.update_system")
        self.assertTrue(ret)



from textwrap import dedent

import jinja2
import pytest
import yaml as _yaml

import salt.serializers.configparser as configparser
import salt.serializers.json as json
import salt.serializers.keyvalue as keyvalue
import salt.serializers.msgpack as msgpack
import salt.serializers.plist as plist
import salt.serializers.python as python
import salt.serializers.tomlmod as tomlmod
import salt.serializers.yaml as yaml
import salt.serializers.yamlex as yamlex
import salt.utils.platform
from salt.serializers import SerializationError
from salt.serializers.yaml import EncryptedString
from salt.utils.odict import OrderedDict

SKIP_MESSAGE = "{} is unavailable, have prerequisites been met?"


@pytest.mark.skipif(json.available is False, reason=SKIP_MESSAGE.format("json"))
def test_serialize_json():
    data = {"foo": "bar"}
    serialized = json.serialize(data)
    assert serialized == '{"foo": "bar"}', serialized

    deserialized = json.deserialize(serialized)
    assert deserialized == data, deserialized


@pytest.mark.skipif(yaml.available is False, reason=SKIP_MESSAGE.format("yaml"))
def test_serialize_yaml():
    data = {"foo": "bar", "encrypted_data": EncryptedString("foo")}
    # The C dumper produces unquoted strings when serializing an
    # EncryptedString, while the non-C dumper produces quoted strings.
    expected = (
        "{encrypted_data: !encrypted foo, foo: bar}"
        if hasattr(_yaml, "CSafeDumper")
        else "{encrypted_data: !encrypted 'foo', foo: bar}"
    )
    serialized = yaml.serialize(data)
    assert serialized == expected, serialized

    deserialized = yaml.deserialize(serialized)
    assert deserialized == data, deserialized


@pytest.mark.skipif(yaml.available is False, reason=SKIP_MESSAGE.format("sls"))
def test_serialize_sls():
    data = {"foo": "bar"}
    serialized = yamlex.serialize(data)
    assert serialized == "{foo: bar}", serialized

    serialized = yamlex.serialize(data, default_flow_style=False)
    assert serialized == "foo: bar", serialized

    deserialized = yamlex.deserialize(serialized)
    assert deserialized == data, deserialized

    serialized = yaml.serialize(data)
    assert serialized == "{foo: bar}", serialized

    deserialized = yaml.deserialize(serialized)
    assert deserialized == data, deserialized

    serialized = yaml.serialize(data, default_flow_style=False)
    assert serialized == "foo: bar", serialized

    deserialized = yaml.deserialize(serialized)
    assert deserialized == data, deserialized


@pytest.mark.skipif(yamlex.available is False, reason=SKIP_MESSAGE.format("sls"))
def test_serialize_complex_sls():
    data = OrderedDict([("foo", 1), ("bar", 2), ("baz", True)])
    serialized = yamlex.serialize(data)
    assert serialized == "{foo: 1, bar: 2, baz: true}", serialized

    deserialized = yamlex.deserialize(serialized)
    assert deserialized == data, deserialized

    serialized = yaml.serialize(data)
    assert serialized == "{bar: 2, baz: true, foo: 1}", serialized

    deserialized = yaml.deserialize(serialized)
    assert deserialized == data, deserialized


@pytest.mark.skipif(yaml.available is False, reason=SKIP_MESSAGE.format("yaml"))
@pytest.mark.skipif(yamlex.available is False, reason=SKIP_MESSAGE.format("sls"))
def test_compare_sls_vs_yaml():
    src = "{foo: 1, bar: 2, baz: {qux: true}}"
    sls_data = yamlex.deserialize(src)
    yml_data = yaml.deserialize(src)

    # ensure that sls & yaml have the same base
    assert isinstance(sls_data, dict)
    assert isinstance(yml_data, dict)
    assert sls_data == yml_data

    # ensure that sls is ordered, while yaml not
    assert isinstance(sls_data, OrderedDict)
    assert not isinstance(yml_data, OrderedDict)


@pytest.mark.skipif(yaml.available is False, reason=SKIP_MESSAGE.format("yaml"))
@pytest.mark.skipif(yamlex.available is False, reason=SKIP_MESSAGE.format("sls"))
def test_compare_sls_vs_yaml_with_jinja():
    tpl = "{{ data }}"
    env = jinja2.Environment()
    src = "{foo: 1, bar: 2, baz: {qux: true}}"

    sls_src = env.from_string(tpl).render(data=yamlex.deserialize(src))
    yml_src = env.from_string(tpl).render(data=yaml.deserialize(src))

    sls_data = yamlex.deserialize(sls_src)
    yml_data = yaml.deserialize(yml_src)

    # ensure that sls & yaml have the same base
    assert isinstance(sls_data, dict)
    assert isinstance(yml_data, dict)
    # The below has been commented out because something the loader test
    # is modifying the yaml renderer to render things to unicode. Without
    # running the loader test, the below passes. Even reloading the module
    # from disk does not reset its internal state (per the Python docs).
    ##
    # assert sls_data == yml_data

    # ensure that sls is ordered, while yaml not
    assert isinstance(sls_data, OrderedDict)
    assert not isinstance(yml_data, OrderedDict)

    # prove that yaml does not handle well with OrderedDict
    # while sls is jinja friendly.
    obj = OrderedDict([("foo", 1), ("bar", 2), ("baz", {"qux": True})])

    sls_obj = yamlex.deserialize(yamlex.serialize(obj))
    try:
        yml_obj = yaml.deserialize(yaml.serialize(obj))
    except SerializationError:
        # BLAAM! yaml was unable to serialize OrderedDict,
        # but it's not the purpose of the current test.
        yml_obj = obj.copy()

    sls_src = env.from_string(tpl).render(data=sls_obj)
    yml_src = env.from_string(tpl).render(data=yml_obj)

    final_obj = yaml.deserialize(sls_src)
    assert obj == final_obj

    # BLAAM! yml_src is not valid !
    final_obj = OrderedDict(yaml.deserialize(yml_src))
    assert obj != final_obj, f"Objects matched! {obj} == {final_obj}"


@pytest.mark.skipif(yamlex.available is False, reason=SKIP_MESSAGE.format("sls"))
def test_sls_aggregate():
    src = dedent(
        """
        a: lol
        foo: !aggregate hello
        bar: !aggregate [1, 2, 3]
        baz: !aggregate
          a: 42
          b: 666
          c: the beast
    """
    ).strip()

    # test that !aggregate is correctly parsed
    sls_obj = yamlex.deserialize(src)
    assert sls_obj == {
        "a": "lol",
        "foo": ["hello"],
        "bar": [1, 2, 3],
        "baz": {"a": 42, "b": 666, "c": "the beast"},
    }, sls_obj

    assert (
        dedent(
            """
        a: lol
        foo: [hello]
        bar: [1, 2, 3]
        baz: {a: 42, b: 666, c: the beast}
    """
        ).strip()
        == yamlex.serialize(sls_obj)
    ), sls_obj

    # test that !aggregate aggregates scalars
    src = dedent(
        """
        placeholder: !aggregate foo
        placeholder: !aggregate bar
        placeholder: !aggregate baz
    """
    ).strip()

    sls_obj = yamlex.deserialize(src)
    assert sls_obj == {"placeholder": ["foo", "bar", "baz"]}, sls_obj

    # test that !aggregate aggregates lists
    src = dedent(
        """
        placeholder: !aggregate foo
        placeholder: !aggregate [bar, baz]
        placeholder: !aggregate []
        placeholder: !aggregate ~
    """
    ).strip()

    sls_obj = yamlex.deserialize(src)
    assert sls_obj == {"placeholder": ["foo", "bar", "baz"]}, sls_obj

    # test that !aggregate aggregates dicts
    src = dedent(
        """
        placeholder: !aggregate {foo: 42}
        placeholder: !aggregate {bar: null}
        placeholder: !aggregate {baz: inga}
    """
    ).strip()

    sls_obj = yamlex.deserialize(src)
    assert sls_obj == {"placeholder": {"foo": 42, "bar": None, "baz": "inga"}}, sls_obj

    # test that !aggregate aggregates deep dicts
    src = dedent(
        """
        placeholder: {foo: !aggregate {foo: 42}}
        placeholder: {foo: !aggregate {bar: null}}
        placeholder: {foo: !aggregate {baz: inga}}
    """
    ).strip()

    sls_obj = yamlex.deserialize(src)
    assert sls_obj == {
        "placeholder": {"foo": {"foo": 42, "bar": None, "baz": "inga"}}
    }, sls_obj

    # test that {foo: !aggregate bar} and {!aggregate foo: bar}
    # are roughly equivalent.
    src = dedent(
        """
        placeholder: {!aggregate foo: {foo: 42}}
        placeholder: {!aggregate foo: {bar: null}}
        placeholder: {!aggregate foo: {baz: inga}}
    """
    ).strip()

    sls_obj = yamlex.deserialize(src)
    assert sls_obj == {
        "placeholder": {"foo": {"foo": 42, "bar": None, "baz": "inga"}}
    }, sls_obj


@pytest.mark.skipif(yamlex.available is False, reason=SKIP_MESSAGE.format("sls"))
def test_sls_reset():
    src = dedent(
        """
        placeholder: {!aggregate foo: {foo: 42}}
        placeholder: {!aggregate foo: {bar: null}}
        !reset placeholder: {!aggregate foo: {baz: inga}}
    """
    ).strip()

    sls_obj = yamlex.deserialize(src)
    assert sls_obj == {"placeholder": {"foo": {"baz": "inga"}}}, sls_obj


@pytest.mark.skipif(yamlex.available is False, reason=SKIP_MESSAGE.format("sls"))
def test_sls_repr():
    """
    Ensure that obj __repr__ and __str__ methods are yaml friendly.
    """

    def convert(obj):
        return yamlex.deserialize(yamlex.serialize(obj))

    sls_obj = convert(OrderedDict([("foo", "bar"), ("baz", "qux")]))

    # ensure that repr and str are yaml friendly
    assert str(sls_obj) == "{foo: bar, baz: qux}"
    assert repr(sls_obj) == "{foo: bar, baz: qux}"

    # ensure that repr and str are already quoted
    assert str(sls_obj["foo"]) == '"bar"'
    assert repr(sls_obj["foo"]) == '"bar"'


@pytest.mark.skipif(yamlex.available is False, reason=SKIP_MESSAGE.format("sls"))
def test_sls_micking_file_merging():
    def convert(obj):
        return yamlex.deserialize(yamlex.serialize(obj))

    # let say that we have 2 pillar files

    src1 = dedent(
        """
        a: first
        b: !aggregate first
        c:
          subkey1: first
          subkey2: !aggregate first
    """
    ).strip()

    src2 = dedent(
        """
        a: second
        b: !aggregate second
        c:
          subkey2: !aggregate second
          subkey3: second
    """
    ).strip()

    sls_obj1 = yamlex.deserialize(src1)
    sls_obj2 = yamlex.deserialize(src2)
    sls_obj3 = yamlex.merge_recursive(sls_obj1, sls_obj2)

    assert sls_obj3 == {
        "a": "second",
        "b": ["first", "second"],
        "c": {"subkey2": ["first", "second"], "subkey3": "second"},
    }, sls_obj3


@pytest.mark.skipif(msgpack.available is False, reason=SKIP_MESSAGE.format("msgpack"))
def test_msgpack():
    data = OrderedDict([("foo", 1), ("bar", 2), ("baz", True)])
    serialized = msgpack.serialize(data)
    deserialized = msgpack.deserialize(serialized)
    assert deserialized == data, deserialized


@pytest.mark.skipif(python.available is False, reason=SKIP_MESSAGE.format("python"))
def test_serialize_python():
    data = {"foo": "bar"}
    serialized = python.serialize(data)
    expected = repr({"foo": "bar"})
    assert serialized == expected, serialized


@pytest.mark.skipif(
    configparser.available is False, reason=SKIP_MESSAGE.format("configparser")
)
def test_configparser():
    data = {"foo": {"bar": "baz"}}
    # configparser appends empty lines
    serialized = configparser.serialize(data).strip()
    assert serialized == "[foo]\nbar = baz", serialized

    deserialized = configparser.deserialize(serialized)
    assert deserialized == data, deserialized


@pytest.mark.skipif(tomlmod.HAS_TOML is False, reason=SKIP_MESSAGE.format("toml"))
def test_serialize_toml():
    data = {"foo": "bar"}
    serialized = tomlmod.serialize(data)
    assert serialized == 'foo = "bar"\n', serialized

    deserialized = tomlmod.deserialize(serialized)
    assert deserialized == data, deserialized


@pytest.mark.skipif(plist.available is False, reason=SKIP_MESSAGE.format("plist"))
def test_serialize_plist():
    data = {"foo": "bar"}
    serialized = plist.serialize(data)
    expected = (
        b'<?xml version="1.0" encoding="UTF-8"?>\n'
        b'<!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN"'
        b' "http://www.apple.com/DTDs/PropertyList-1.0.dtd">\n'
        b'<plist version="1.0">\n'
        b"<dict>\n"
        b"\t<key>foo</key>\n"
        b"\t<string>bar</string>\n"
        b"</dict>\n"
        b"</plist>\n"
    )
    assert serialized == expected, serialized

    deserialized = plist.deserialize(serialized)
    assert deserialized == data, deserialized


@pytest.mark.skipif(plist.available is False, reason=SKIP_MESSAGE.format("plist"))
def test_serialize_binary_plist():
    data = {"foo": "bar"}
    serialized = plist.serialize(data, fmt="FMT_BINARY")

    deserialized = plist.deserialize(serialized)
    assert deserialized == data, deserialized


def test_serialize_keyvalue():
    data = {"foo": "bar baz"}
    serialized = keyvalue.serialize(data)
    assert serialized == "foo=bar baz", serialized

    deserialized = keyvalue.deserialize(serialized)
    assert deserialized == data, deserialized


def test_serialize_keyvalue_quoting():
    data = {"foo": "bar baz"}
    serialized = keyvalue.serialize(data, quoting=True)
    assert serialized == "foo='bar baz'", serialized

    deserialized = keyvalue.deserialize(serialized, quoting=False)
    assert deserialized == data, deserialized


def test_serialize_keyvalue_separator():
    data = {"foo": "bar baz"}
    serialized = keyvalue.serialize(data, separator=" = ")
    assert serialized == "foo = bar baz", serialized

    deserialized = keyvalue.deserialize(serialized, separator=" = ")
    assert deserialized == data, deserialized


def test_serialize_keyvalue_list_of_lists():
    if salt.utils.platform.is_windows():
        linend = "\r\n"
    else:
        linend = "\n"
    data = [["foo", "bar baz"], ["salt", "rocks"]]
    expected = {"foo": "bar baz", "salt": "rocks"}
    serialized = keyvalue.serialize(data)
    assert serialized == f"foo=bar baz{linend}salt=rocks", serialized

    deserialized = keyvalue.deserialize(serialized)
    assert deserialized == expected, deserialized



"""
SCSI administration module
"""

import logging
import os.path

import salt.utils.path

log = logging.getLogger(__name__)

__func_alias__ = {"ls_": "ls"}


def ls_(get_size=True):
    """
    List SCSI devices, with details

    CLI Examples:

    .. code-block:: bash

        salt '*' scsi.ls
        salt '*' scsi.ls get_size=False

    get_size : True
        Get the size information for scsi devices.  This option
        should be set to False for older OS distributions (RHEL6 and older)
        due to lack of support for the '-s' option in lsscsi.

        .. versionadded:: 2015.5.10
    """

    if not salt.utils.path.which("lsscsi"):
        __context__["retcode"] = 1
        return "scsi.ls not available - lsscsi command not found"

    if get_size:
        cmd = "lsscsi -dLsv"
    else:
        cmd = "lsscsi -dLv"

    ret = {}
    res = __salt__["cmd.run_all"](cmd)
    rc = res.get("retcode", -1)
    if rc != 0:
        __context__["retcode"] = rc
        error = res.get("stderr", "").split("\n")[0]
        if error == "lsscsi: invalid option -- 's'":
            return f"{error} - try get_size=False"
        return res.get("stderr", "").split("\n")[0]
    data = res.get("stdout", "")

    for line in data.splitlines():
        if line.startswith("["):
            size = None
            major = None
            minor = None
            comps = line.strip().split()
            key = comps[0]
            if get_size:
                size = comps.pop()
            majmin = comps.pop()
            if majmin.startswith("["):
                major, minor = majmin.replace("[", "").replace("]", "").split(":")
            device = comps.pop()
            model = " ".join(comps[3:])
            ret[key] = {
                "lun": key.replace("[", "").replace("]", ""),
                "size": size,
                "major": major,
                "minor": minor,
                "device": device,
                "model": model,
            }
        elif line.startswith(" "):
            if line.strip().startswith("dir"):
                comps = line.strip().split()
                ret[key]["dir"] = [comps[1], comps[2].replace("[", "").replace("]", "")]
            else:
                comps = line.strip().split("=")
                ret[key][comps[0]] = comps[1]
    return ret


def rescan_all(host):
    """
    List scsi devices

    CLI Example:

    .. code-block:: bash

        salt '*' scsi.rescan_all 0
    """
    if os.path.isdir(f"/sys/class/scsi_host/host{host}"):
        cmd = f'echo "- - -" > /sys/class/scsi_host/host{host}/scan'
    else:
        return f"Host {host} does not exist"
    return __salt__["cmd.run"](cmd).splitlines()



import logging
import os
import shutil
import textwrap

import pytest

import salt.config
import salt.loader
import salt.modules.cmdmod as cmdmod
import salt.modules.config as configmod
import salt.modules.file as filemod
import salt.utils.data
import salt.utils.files
import salt.utils.platform
import salt.utils.stringutils
from salt.exceptions import SaltInvocationError
from tests.support.mock import MagicMock

log = logging.getLogger(__name__)

pytestmark = pytest.mark.skipif(
    salt.utils.platform.is_windows(), reason="grep not supported on Windows"
)


@pytest.fixture
def configure_loader_modules():
    return {
        filemod: {
            "__salt__": {
                "config.manage_mode": configmod.manage_mode,
                "cmd.run": cmdmod.run,
                "cmd.run_all": cmdmod.run_all,
            },
            "__opts__": {
                "test": False,
                "file_roots": {"base": "tmp"},
                "pillar_roots": {"base": "tmp"},
                "cachedir": "tmp",
                "grains": {},
            },
            "__grains__": {"kernel": "Linux"},
            "__utils__": {
                "files.is_text": MagicMock(return_value=True),
                "stringutils.get_diff": salt.utils.stringutils.get_diff,
            },
        }
    }


@pytest.fixture
def multiline_string():
    multiline_string = textwrap.dedent(
        """\
        Lorem ipsum dolor sit amet, consectetur
        adipiscing elit. Nam rhoncus enim ac
        bibendum vulputate.
        """
    )

    multiline_string = os.linesep.join(multiline_string.splitlines())

    return multiline_string


@pytest.fixture
def multiline_file(tmp_path, multiline_string):
    multiline_file = str(tmp_path / "multiline-file.txt")

    with salt.utils.files.fopen(multiline_file, "w+") as file_handle:
        file_handle.write(multiline_string)

    yield multiline_file
    shutil.rmtree(str(tmp_path))


def test_grep_query_exists(multiline_file):
    result = filemod.grep(multiline_file, "Lorem ipsum")

    assert result, None
    assert result["retcode"] == 0
    assert result["stdout"] == "Lorem ipsum dolor sit amet, consectetur"
    assert result["stderr"] == ""


def test_grep_query_not_exists(multiline_file):
    result = filemod.grep(multiline_file, "Lorem Lorem")

    assert result["retcode"] == 1
    assert result["stdout"] == ""
    assert result["stderr"] == ""


def test_grep_query_exists_with_opt(multiline_file):
    result = filemod.grep(multiline_file, "Lorem ipsum", "-i")

    assert result, None
    assert result["retcode"] == 0
    assert result["stdout"] == "Lorem ipsum dolor sit amet, consectetur"
    assert result["stderr"] == ""


def test_grep_query_not_exists_opt(multiline_file, multiline_string):
    result = filemod.grep(multiline_file, "Lorem Lorem", "-v")

    assert result["retcode"] == 0
    assert result["stdout"] == multiline_string
    assert result["stderr"] == ""


def test_grep_query_too_many_opts(multiline_file):
    with pytest.raises(
        SaltInvocationError, match="^Passing multiple command line arg"
    ) as cm:
        result = filemod.grep(multiline_file, "Lorem Lorem", "-i -b2")


def test_grep_query_exists_wildcard(multiline_file):
    _file = f"{multiline_file}*"
    result = filemod.grep(_file, "Lorem ipsum")

    assert result, None
    assert result["retcode"] == 0
    assert result["stdout"] == "Lorem ipsum dolor sit amet, consectetur"
    assert result["stderr"] == ""


def test_grep_file_not_exists_wildcard(multiline_file):
    _file = f"{multiline_file}-junk*"
    result = filemod.grep(_file, "Lorem ipsum")

    assert result, None
    assert not result["retcode"] == 0
    assert not result["stdout"] == "Lorem ipsum dolor sit amet, consectetur"
    _expected_stderr = "grep: {}-junk*: No such file or directory".format(
        multiline_file
    )
    assert result["stderr"] == _expected_stderr



import pytest

pytestmark = [
    pytest.mark.core_test,
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
]


@pytest.fixture(scope="module")
def account():
    with pytest.helpers.create_account() as _account:
        yield _account


@pytest.mark.parametrize(
    "exit_code, return_code, result",
    [
        (300, 0, True),
        (299, 299, False),
    ],
)
def test_script_exitcode(modules, state_tree, exit_code, return_code, result):
    ret = modules.state.single(
        "cmd.run", name=f"cmd.exe /c exit {exit_code}", success_retcodes=[2, 44, 300]
    )
    assert ret.result is result
    assert ret.filtered["changes"]["retcode"] == return_code


@pytest.mark.parametrize(
    "exit_code, return_code, result",
    [
        (300, 0, True),
        (299, 299, False),
    ],
)
def test_script_exitcode_runas(
    modules, state_tree, exit_code, return_code, result, account
):
    ret = modules.state.single(
        "cmd.run",
        name=f"cmd.exe /c exit {exit_code}",
        success_retcodes=[2, 44, 300],
        runas=account.username,
        password=account.password,
    )
    assert ret.result is result
    assert ret.filtered["changes"]["retcode"] == return_code


@pytest.mark.parametrize(
    "command, expected",
    [
        ("echo foo", "foo"),
        ("cmd /c echo foo", "foo"),
        ("whoami && echo foo", "foo"),
        ("cmd /c whoami && echo foo", "foo"),
    ],
)
def test_run_builtins(modules, command, expected):
    result = modules.cmd.run(command)
    assert expected in result


@pytest.mark.parametrize(
    "command, expected",
    [
        ("echo foo", "foo"),
        ("cmd /c echo foo", "foo"),
        ("whoami && echo foo", "foo"),
        ("cmd /c whoami && echo foo", "foo"),
    ],
)
def test_run_builtins_runas(modules, account, command, expected):
    result = modules.cmd.run(
        cmd=command, runas=account.username, password=account.password
    )
    assert expected in result



import logging

import pytest

import salt.utils.platform
import salt.utils.user

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.destructive_test,
    pytest.mark.skip_if_not_root,
    pytest.mark.skip_on_windows,
]


@pytest.fixture(scope="module")
def user():
    with pytest.helpers.create_account(create_group=True) as _account:
        yield _account


@pytest.fixture(scope="module")
def dupegroup(user):
    grpid = user.group.info.gid
    with pytest.helpers.create_group(
        name="dupegroup", gid=grpid, members=user.username
    ) as _group:
        yield _group


@pytest.mark.skip_on_platforms(
    darwin=True,
    freebsd=True,
    reason="This test should not run on FreeBSD and Mac due to lack of duplicate GID support",
)
def test_get_group_list_with_duplicate_gid_group(user, dupegroup):
    group_list = salt.utils.user.get_group_list(user.username)
    assert user.group.info.name in group_list
    assert dupegroup.name in group_list



"""test for pillar stack pillar"""

import pytest

import salt.pillar.stack as stack
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    loader_globals = {
        "__grains__": {"os": "Debian", "os_family": "Debian"},
        "__opts__": {"saltenv": "dev", "pillarenv": "dev"},
    }
    return {stack: loader_globals}


def mock_stack_pillar(mock_output, *args, **kwargs):
    # mock: jenv.get_template(filename).render(stack=stack)
    class MockJinja:
        def __call__(self, *args, **kwargs):
            return self

        render = MagicMock(side_effect=mock_output)

    with patch("os.path.isfile", MagicMock(return_value=True)), patch(
        "jinja2.environment.Environment.get_template", MockJinja()
    ), patch("glob.glob", MagicMock(return_value=["/path/to/stack.cfg"])):
        result = stack.ext_pillar(  # (minion_id, pillar, *args, **kwargs)
            "minion_id", {}, *args, **kwargs
        )
        return result


def test_extpillar_stack1():

    mock_output = [
        "/path/to/filename.yml\n",  # mocked contents of /path/to/stack.cfg
        """
            foo: foo1 # jinja test
            bar: bar1
        """,  # mocked contents of filename.yml
    ]
    fake_dict = {"foo": "foo1", "bar": "bar1"}

    # config with a single file
    result = mock_stack_pillar(mock_output, "/path/to/stack.cfg")
    assert fake_dict == result

    # config with a opts:saltenv
    result = mock_stack_pillar(
        mock_output,
        **{
            "opts:saltenv": {  # **kwargs
                "dev": "/path/to/dev/static.cfg",
            }
        }
    )
    assert fake_dict == result

    # config with a opts:saltenv and __env__ substitution
    result = mock_stack_pillar(
        mock_output,
        **{
            "opts:saltenv": {  # **kwargs
                "__env__": "/path/to/__env__/dynamic.cfg",
            }
        }
    )
    assert fake_dict == result


def test_extpillar_stack_exceptions():

    # yaml indentation error
    mock_output = [
        "/path/to/filename.yml\n",  # mocked contents of /path/to/stack.cfg
        """
                foo: foo1
            bar: bar1  # yaml indentation error
        """,  # mocked contents of filename.yml
    ]
    pytest.raises(Exception, mock_stack_pillar, mock_output, "/path/to/stack.cfg")
